import torch
from abc import abstractmethod
from torch import nn
from torch_scatter import scatter
from typing import Dict, List, Tuple


class AbstractMessagePassingLayer(nn.Module):
    """Interface for message passing layer to be used in graph neural networks with multiple edge types."""

    @abstractmethod
    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        """
        :param node_states: A [num_nodes, D] matrix containing the states of all nodes.
        :param adjacency_lists: A list of [num_edges, 2] adjacency lists, one per edge type.
        :param node_to_graph_idx:
        :param reference_node_ids:
        :param reference_node_graph_idx:
        :param edge_features: A list of [num_edges, H] with edge features.
            Has the size of `adjacency_lists`.
        :return: the next node states in a [num_nodes, D'] matrix.
        """

    def _aggregate_messages(
        self, messages: torch.Tensor, message_targets: torch.Tensor, num_nodes, aggregation_fn: str
    ):
        """Utility function to be used by concrete implementors."""
        # Support AMP
        msg_dtype = messages.dtype
        return scatter(
            messages.to(torch.float32),
            index=message_targets,
            dim=0,
            dim_size=num_nodes,
            reduce=aggregation_fn,
        ).to(msg_dtype)

    @property
    @abstractmethod
    def input_state_dimension(self) -> int:
        pass

    @property
    @abstractmethod
    def output_state_dimension(self) -> int:
        pass


class AbstractMessageAggregation(nn.Module):
    @abstractmethod
    def forward(self, messages: torch.Tensor, message_targets: torch.Tensor, num_nodes):
        pass

    @abstractmethod
    def output_state_size(self, message_input_size: int) -> int:
        pass



from typing_extensions import Literal

import torch
from abc import abstractmethod
from math import sqrt
from torch import nn
from torch_scatter import scatter, scatter_log_softmax, scatter_sum
from typing import NamedTuple, Union


class ElementsToSummaryRepresentationInput(NamedTuple):
    """Input to AbstractVarSizedElementReduce layers."""

    element_embeddings: torch.Tensor  # float tensor of shape [num_elements, D], the representation of each node in all graphs.
    element_to_sample_map: torch.Tensor  # int tensor of shape [num_elements] with values in range [0, num_sampless-1], mapping each node to a sample ID.

    num_samples: Union[torch.Tensor, int]  # scalar, specifying the number of sets.


class AbstractVarSizedElementReduce(nn.Module):
    """Interface for computing summary representations from multiple variable-sized sets of representations."""

    @abstractmethod
    def forward(self, inputs: ElementsToSummaryRepresentationInput) -> torch.Tensor:
        """Returns: float tensor of shape [num_samples, D']"""


class SimpleVarSizedElementReduce(AbstractVarSizedElementReduce):
    def __init__(self, summarization_type: Literal["sum", "mean", "max", "min"]):
        super().__init__()
        assert summarization_type in {"sum", "mean", "max", "min"}
        self.__summarization_type = summarization_type

    def forward(self, inputs: ElementsToSummaryRepresentationInput) -> torch.Tensor:
        return scatter(
            src=inputs.element_embeddings,
            index=inputs.element_to_sample_map,
            dim=0,
            dim_size=inputs.num_samples,
            reduce=self.__summarization_type,
        )


class NormalizedWeightsVarSizedElementReduce(AbstractVarSizedElementReduce):
    def __init__(self, input_representation_size: int, output_representation_size: int):
        super().__init__()
        self.__attention_layer = nn.Linear(input_representation_size, 1, bias=False)
        self.__output_layer = nn.Linear(
            input_representation_size, output_representation_size, bias=False
        )

    def forward(self, inputs: ElementsToSummaryRepresentationInput) -> torch.Tensor:
        attention_scores = self.__attention_layer(inputs.element_embeddings).squeeze(
            -1
        )  # [num_elements]
        attention_probs = torch.exp(
            scatter_log_softmax(attention_scores, index=inputs.element_to_sample_map, dim=0, eps=0)
        )  # [num_elements]
        return scatter_sum(
            self.__output_layer(inputs.element_embeddings) * attention_probs.unsqueeze(-1),
            index=inputs.num_samples,
            dim=0,
            dim_size=inputs.num_samples,
        )  # [num_samples, D']


class WeightedSumVarSizedElementReduce(AbstractVarSizedElementReduce):
    def __init__(self, representation_size: int):
        super().__init__()
        self.__weights_layer = nn.Linear(representation_size, 1, bias=False)

    def forward(self, inputs: ElementsToSummaryRepresentationInput) -> torch.Tensor:
        weights = torch.sigmoid(
            self.__weights_layer(inputs.element_embeddings).squeeze(-1)
        )  # [num_elements]
        return scatter_sum(
            inputs.element_embeddings * weights.unsqueeze(-1),
            index=inputs.element_to_sample_map,
            dim=0,
            dim_size=inputs.num_samples,
        )  # [num_samples, D']


class SelfAttentionVarSizedElementReduce(AbstractVarSizedElementReduce):
    def __init__(
        self,
        input_representation_size: int,
        hidden_size: int,
        output_representation_size: int,
        query_representation_summarizer: AbstractVarSizedElementReduce,
    ):
        super().__init__()
        self.__query_layer = query_representation_summarizer
        self.__key_layer = nn.Linear(input_representation_size, hidden_size, bias=False)
        self.__output_layer = nn.Linear(
            input_representation_size, output_representation_size, bias=False
        )

    def forward(self, inputs: ElementsToSummaryRepresentationInput) -> torch.Tensor:
        queries = self.__query_layer(inputs)  # [num_samples, H]
        queries_all = queries[inputs.element_to_sample_map]  # [num_elements, H]
        keys = self.__key_layer(inputs.element_embeddings)  # [num_elements, H]

        attention_scores = torch.einsum("vh,vh->v", queries_all, keys)  # [num_elements]
        attention_probs = torch.exp(
            scatter_log_softmax(attention_scores, index=inputs.element_to_sample_map, dim=0, eps=0)
        )  # [num_elements]
        return scatter_sum(
            self.__output_layer(inputs.element_embeddings) * attention_probs.unsqueeze(-1),
            index=inputs.element_to_sample_map,
            dim=0,
            dim_size=inputs.num_samples,
        )  # [num_samples, D']


class MultiheadSelfAttentionVarSizedElementReduce(AbstractVarSizedElementReduce):
    def __init__(
        self,
        input_representation_size: int,
        hidden_size: int,
        output_representation_size: int,
        num_heads: int,
        query_representation_summarizer: AbstractVarSizedElementReduce,
        use_value_layer: bool = False,
    ):
        super().__init__()
        self.__query_layer = query_representation_summarizer
        self.__key_layer = nn.Linear(input_representation_size, hidden_size, bias=False)
        assert hidden_size % num_heads == 0, "Hidden size must be divisible by the number of heads."
        self.__use_value_layer = use_value_layer
        if use_value_layer:
            self.__value_layer = nn.Linear(input_representation_size, hidden_size, bias=False)
            self.__output_layer = nn.Linear(hidden_size, output_representation_size, bias=False)
        else:
            self.__output_layer = nn.Linear(
                input_representation_size * num_heads, output_representation_size, bias=False
            )
        self.__num_heads = num_heads

    def forward(self, inputs: ElementsToSummaryRepresentationInput) -> torch.Tensor:
        queries = self.__query_layer(inputs)  # [num_samples, H]
        queries_per_element = queries[inputs.element_to_sample_map]  # [num_elements, H]
        queries_per_element = queries_per_element.reshape(
            (
                queries_per_element.shape[0],
                self.__num_heads,
                queries_per_element.shape[1] // self.__num_heads,
            )
        )

        keys = self.__key_layer(inputs.element_embeddings)  # [num_elements, H]
        keys = keys.reshape((keys.shape[0], self.__num_heads, keys.shape[1] // self.__num_heads))

        attention_scores = torch.einsum("bkh,bkh->bk", queries_per_element, keys) / sqrt(
            keys.shape[-1]
        )  # [num_elements, num_heads]
        attention_probs = torch.exp(
            scatter_log_softmax(attention_scores, index=inputs.element_to_sample_map, dim=0, eps=0)
        )  # [num_elements, num_heads]

        if self.__use_value_layer:
            values = self.__value_layer(inputs.element_embeddings)  # [num_elements, hidden_size]
            values = values.reshape(
                (values.shape[0], self.__num_heads, values.shape[1] // self.__num_heads)
            )
            outputs = attention_probs.unsqueeze(-1) * values
        else:
            outputs = attention_probs.unsqueeze(-1) * inputs.element_embeddings.unsqueeze(
                1
            )  # [num_elements, num_heads, D']

        outputs = outputs.reshape((outputs.shape[0], -1))  # [num_elements, num_heads * D']

        per_sample_outputs = scatter_sum(
            outputs, index=inputs.element_to_sample_map, dim=0, dim_size=inputs.num_samples
        )  # [num_samples, num_heads, D']

        return self.__output_layer(per_sample_outputs)  # [num_samples, D']



#!/usr/bin/env python
"""
Usage:
    train.py [options] TRAIN_DATA_PATH VALID_DATA_PATH TEST_DATA_PATH MODEL_FILENAME

Options:
    --aml                      Run this in Azure ML
    --amp                      Enable automatic mixed precision.
    --azure-info=<path>        Azure authentication information file (JSON). Used to load data from Azure storage.
    --max-num-epochs=<epochs>  The maximum number of epochs to run training for. [default: 100]
    --minibatch-size=<size>    The minibatch size. [default: 300]
    --restore-path=<path>      The path to previous model file for starting from previous checkpoint.
    --sequential-run           Do not parallelize data loading. Makes debugging easier.
    --quiet                    Do not show progress bar.
    -h --help                  Show this screen.
    --debug                    Enable debug routines. [default: False]
"""
import random
import torch
from docopt import docopt
from dpu_utils.utils import RichPath, run_and_debug
from pathlib import Path

from ptgnn.baseneuralmodel import ModelTrainer
from ptgnn.baseneuralmodel.utils.amlutils import configure_logging, log_run
from ptgnn.baseneuralmodel.utils.data import LazyDataIterable
from ptgnn.implementations.typilus.graph2class import Graph2Class
from ptgnn.neuralmodels.embeddings.strelementrepresentationmodel import (
    StrElementRepresentationModel,
)
from ptgnn.neuralmodels.gnn import GraphNeuralNetworkModel
from ptgnn.neuralmodels.gnn.messagepassing import GatedMessagePassingLayer, MlpMessagePassingLayer
from ptgnn.neuralmodels.gnn.messagepassing.residuallayers import ConcatResidualLayer


def create_graph2class_gnn_model(hidden_state_size: int = 64, dropout_rate: float = 0.1):
    def create_ggnn_mp_layers(num_edges: int):
        ggnn_mp = GatedMessagePassingLayer(
            state_dimension=hidden_state_size,
            message_dimension=hidden_state_size,
            num_edge_types=num_edges,
            message_aggregation_function="max",
            dropout_rate=dropout_rate,
        )
        r1 = ConcatResidualLayer(hidden_state_size)
        return [
            r1.pass_through_dummy_layer(),
            ggnn_mp,
            ggnn_mp,
            ggnn_mp,
            ggnn_mp,
            ggnn_mp,
            ggnn_mp,
            ggnn_mp,
            r1,
            GatedMessagePassingLayer(
                state_dimension=2 * hidden_state_size,
                message_dimension=hidden_state_size,
                num_edge_types=num_edges,
                message_aggregation_function="max",
                dropout_rate=dropout_rate,
            ),
        ]

    def create_mlp_mp_layers(num_edges: int):
        mlp_mp_constructor = lambda: MlpMessagePassingLayer(
            input_state_dimension=hidden_state_size,
            message_dimension=hidden_state_size,
            output_state_dimension=hidden_state_size,
            num_edge_types=num_edges,
            message_aggregation_function="max",
            dropout_rate=0.1,
        )
        mlp_mp_after_res_constructor = lambda: MlpMessagePassingLayer(
            input_state_dimension=2 * hidden_state_size,
            message_dimension=2 * hidden_state_size,
            output_state_dimension=hidden_state_size,
            num_edge_types=num_edges,
            message_aggregation_function="max",
            dropout_rate=0.1,
        )
        r1 = ConcatResidualLayer(hidden_state_size)
        r2 = ConcatResidualLayer(hidden_state_size)
        return [
            r1.pass_through_dummy_layer(),
            mlp_mp_constructor(),
            mlp_mp_constructor(),
            mlp_mp_constructor(),
            r1,
            mlp_mp_after_res_constructor(),
            r2.pass_through_dummy_layer(),
            mlp_mp_constructor(),
            mlp_mp_constructor(),
            mlp_mp_constructor(),
            r2,
            mlp_mp_after_res_constructor(),
        ]

    return Graph2Class(
        gnn_model=GraphNeuralNetworkModel(
            node_representation_model=StrElementRepresentationModel(
                embedding_size=hidden_state_size,
                token_splitting="subtoken",
                subtoken_combination="mean",
                vocabulary_size=10000,
                min_freq_threshold=5,
                dropout_rate=dropout_rate,
            ),
            message_passing_layer_creator=create_mlp_mp_layers,
            max_nodes_per_graph=100000,
            max_graph_edges=500000,
            introduce_backwards_edges=True,
            add_self_edges=True,
            stop_extending_minibatch_after_num_nodes=120000,
            edge_dropout_rate=0.0,
        ),
        max_num_classes=100,
    )


def load_from_folder(path: RichPath, shuffle: bool):
    all_files = path.get_filtered_files_in_dir("*.jsonl.gz")
    if shuffle:
        random.shuffle(all_files)
    for file in all_files:
        yield from file.read_as_jsonl()


def run(arguments):
    if arguments["--aml"]:
        from azureml.core.run import Run

        aml_ctx = Run.get_context()
        assert torch.cuda.is_available(), "No CUDA available. Aborting training."
    else:
        aml_ctx = None

    log_path = configure_logging(aml_ctx)
    azure_info_path = arguments.get("--azure-info", None)
    training_data_path = RichPath.create(arguments["TRAIN_DATA_PATH"], azure_info_path)
    training_data = LazyDataIterable(lambda: load_from_folder(training_data_path, shuffle=True))

    validation_data_path = RichPath.create(arguments["VALID_DATA_PATH"], azure_info_path)
    validation_data = LazyDataIterable(
        lambda: load_from_folder(validation_data_path, shuffle=False)
    )

    model_path = Path(arguments["MODEL_FILENAME"])
    assert model_path.name.endswith(".pkl.gz"), "MODEL_FILENAME must have a `.pkl.gz` suffix."

    initialize_metadata = True
    restore_path = arguments.get("--restore-path", None)
    if restore_path:
        initialize_metadata = False
        model, nn = Graph2Class.restore_model(Path(restore_path))
    elif arguments["--aml"] and model_path.exists():
        initialize_metadata = False
        model, nn = Graph2Class.restore_model(model_path)
    else:
        nn = None
        model = create_graph2class_gnn_model()

    def create_optimizer(parameters):
        return torch.optim.Adam(parameters, lr=0.00025)

    trainer = ModelTrainer(
        model,
        model_path,
        max_num_epochs=int(arguments["--max-num-epochs"]),
        minibatch_size=int(arguments["--minibatch-size"]),
        optimizer_creator=create_optimizer,
        clip_gradient_norm=1,
        target_validation_metric="Accuracy",
        target_validation_metric_higher_is_better=True,
        enable_amp=arguments["--amp"],
    )
    if nn is not None:
        trainer.neural_module = nn

    trainer.register_train_epoch_end_hook(
        lambda model, nn, epoch, metrics: log_run(aml_ctx, "train", model, epoch, metrics)
    )
    trainer.register_validation_epoch_end_hook(
        lambda model, nn, epoch, metrics: log_run(aml_ctx, "valid", model, epoch, metrics)
    )

    trainer.train(
        training_data,
        validation_data,
        show_progress_bar=not arguments["--quiet"],
        initialize_metadata=initialize_metadata,
        parallelize=not arguments["--sequential-run"],
        patience=10,
        store_tensorized_data_in_memory=True,
    )

    test_data_path = RichPath.create(arguments["TEST_DATA_PATH"], azure_info_path)
    test_data = LazyDataIterable(lambda: load_from_folder(test_data_path, shuffle=False))
    acc = model.report_accuracy(
        test_data,
        trainer.neural_module,
        device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
    )
    print(f"Test accuracy: {acc:%}")

    if aml_ctx is not None:
        aml_ctx.log("Test Accuracy", acc)
        aml_ctx.upload_file(name="model.pkl.gz", path_or_stream=str(model_path))
        aml_ctx.upload_file(name="full.log", path_or_stream=log_path)


if __name__ == "__main__":
    args = docopt(__doc__)
    run_and_debug(lambda: run(args), args.get("--debug", False))



import numpy as np
import torch
from torch import nn
from typing import Any, Dict, Iterator, NamedTuple, Optional, Union

from ptgnn.baseneuralmodel import AbstractNeuralModel, ModuleWithMetrics
from ptgnn.baseneuralmodel.utils.data import enforce_not_None
from ptgnn.tests.simplemodel.data import SampleDatapoint


class TensorizedDatapoint(NamedTuple):
    input_features: np.ndarray
    target_class: np.ndarray


class SimpleRegressionNetwork(ModuleWithMetrics):
    def __init__(self, num_features: int):
        super().__init__()
        self.__layer = nn.Linear(num_features, 1, bias=True)
        self.__loss = nn.BCEWithLogitsLoss()

    def forward(self, inputs, targets):
        predicted = self.__layer(inputs)[:, 0]  # B
        loss = self.__loss(input=predicted, target=targets)
        return loss

    def predict(self, inputs: torch.Tensor, **kwargs):
        predicted = self.__layer(inputs)[:, 0]  # B
        return predicted >= 0


class SimpleRegressionModel(
    AbstractNeuralModel[SampleDatapoint, TensorizedDatapoint, SimpleRegressionNetwork]
):
    """A simple linear regression model used for testing."""

    def initialize_metadata(self) -> None:
        self.__num_features: Optional[int] = None

    def update_metadata_from(self, datapoint: SampleDatapoint) -> None:
        if self.__num_features is None:
            self.__num_features = len(datapoint.input_features)
        else:
            assert self.__num_features == len(datapoint.input_features)

    def build_neural_module(self) -> SimpleRegressionNetwork:
        return SimpleRegressionNetwork(enforce_not_None(self.__num_features))

    def tensorize(self, datapoint: SampleDatapoint) -> TensorizedDatapoint:
        return TensorizedDatapoint(
            input_features=np.array(datapoint.input_features, dtype=np.float32),
            target_class=np.array(1 if datapoint.target_class else 0, dtype=np.float32),
        )

    def initialize_minibatch(self) -> Dict[str, Any]:
        return {"inputs": [], "targets": []}

    def extend_minibatch_with(
        self, tensorized_datapoint: TensorizedDatapoint, partial_minibatch: Dict[str, Any]
    ) -> bool:
        partial_minibatch["inputs"].append(tensorized_datapoint.input_features)
        partial_minibatch["targets"].append(tensorized_datapoint.target_class)
        return True

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        return {
            "inputs": torch.tensor(
                np.stack(accumulated_minibatch_data["inputs"], axis=0), device=device
            ),
            "targets": torch.tensor(
                np.stack(accumulated_minibatch_data["targets"], axis=0), device=device
            ),
        }

    def compute_accuracy(
        self,
        trained_network: SimpleRegressionNetwork,
        dataset: Iterator[SampleDatapoint],
        parallelize: bool = True,
        use_multiprocessing=True,
    ) -> float:
        trained_network.eval()
        num_samples, num_correct = 0, 0

        for mb_data, initial_points in self.minibatch_iterator(
            self.tensorize_dataset(
                dataset,
                return_input_data=True,
                parallelize=parallelize,
                use_multiprocessing=use_multiprocessing,
            ),
            max_minibatch_size=10,
            parallelize=parallelize,
            device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
        ):
            mb_predictions = trained_network.predict(**mb_data).cpu().numpy()
            for point, prediction in zip(initial_points, mb_predictions):
                num_samples += 1
                if point is None:
                    raise Exception("Inital Point Data was not loaded.")
                if point.target_class == prediction:
                    num_correct += 1
        return num_correct / num_samples



import math
import torch
from torch import nn


class LuongAttentionModule(nn.Module):
    """
    A Luong-style attention that also includes the inner product of targets-lookup.
    """

    def __init__(
        self, memories_hidden_dimension: int, lookup_hidden_dimension: int, output_size: int
    ) -> None:
        super().__init__()
        self.__Whd = nn.Parameter(
            torch.randn(
                memories_hidden_dimension,
                lookup_hidden_dimension,
                dtype=torch.float,
                requires_grad=True,
            )
        )
        self.__Wout = nn.Linear(
            memories_hidden_dimension + lookup_hidden_dimension, output_size, bias=False
        )

    def forward(
        self, *, memories: torch.Tensor, memories_length: torch.Tensor, lookup_vectors: torch.Tensor
    ) -> torch.Tensor:
        # memories: [B, max-inp-len, H]
        # memories_length: [B]
        # look_up_vectors: [B, max-out-len, D]
        return self.forward_with_attention_vec(
            memories=memories, memories_length=memories_length, lookup_vectors=lookup_vectors
        )[0]

    def forward_with_attention_vec(
        self, *, memories: torch.Tensor, memories_length: torch.Tensor, lookup_vectors: torch.Tensor
    ):
        attention = self.get_attention_vector(
            lookup_vectors, memories, memories_length
        )  # [B, max-out-len, max-inp-len]

        contexts = torch.einsum("blq,bqh->blh", attention, memories)  # [B, max-out-len, H]
        hc = torch.cat((contexts, lookup_vectors), dim=-1)  # [B, max-out-len, H]
        return torch.tanh(self.__Wout(hc)), attention

    def get_attention_vector(
        self, lookup_vectors: torch.Tensor, memories: torch.Tensor, memories_length: torch.Tensor
    ) -> torch.Tensor:
        # memories: [B, max-inp-len, H]
        # memories_length: [B]
        # look_up_vectors: [B, max-out-len, D]
        # Output: [B, max-out-len, max-inp-len]
        memories_in_d = torch.einsum("blh,hd->bld", memories, self.__Whd)  # [B, max-inp-len, D]
        logits = torch.einsum(
            "bld,bqd->bql", memories_in_d, lookup_vectors
        )  # [B, max-out-len, max-inp-len]

        mask = (
            torch.arange(memories.shape[1], device=self.device).view(1, -1)
            >= memories_length.view(-1, 1)
        ).unsqueeze(
            1
        )  # [B, 1, max-inp-len]
        logits.masked_fill_(mask, -math.inf)
        attention = nn.functional.softmax(logits, dim=-1)  # [B, max-len]
        return attention



from typing_extensions import Final, TypedDict

import re
import torch
from dpu_utils.codeutils import split_identifier_into_parts
from itertools import chain
from torch import nn
from torch_scatter import scatter_log_softmax, scatter_max
from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Tuple, Union

from ptgnn.baseneuralmodel import AbstractNeuralModel, ModuleWithMetrics
from ptgnn.neuralmodels.gnn.graphneuralnetwork import GraphNeuralNetwork, GraphNeuralNetworkModel
from ptgnn.neuralmodels.gnn.structs import GnnOutput, GraphData, TensorizedGraphData


class VarMisuseGraph(TypedDict):
    Edges: Dict[str, List[Tuple[int, int]]]
    NodeLabels: Dict[str, str]
    NodeTypes: Dict[str, str]


class SymbolCandidate(TypedDict):
    SymbolDummyNode: int
    SymbolName: str
    IsCorrect: bool


class VarMisuseSample(TypedDict):
    ContextGraph: VarMisuseGraph
    slotTokenIdx: int
    SlotDummyNode: int
    SymbolCandidates: List[SymbolCandidate]


class TensorizedVarMisuseSample(NamedTuple):
    graph: TensorizedGraphData
    target_candidate_node_idx: int
    num_candidate_nodes: int


class VarMisuseGraphModel(ModuleWithMetrics):
    def __init__(self, gnn: GraphNeuralNetwork):
        super().__init__()
        self._gnn = gnn
        self.__candidate_scores = nn.Linear(
            self._gnn.output_node_state_dim + self._gnn.output_node_state_dim,
            1,
            bias=False,
        )

    def _reset_module_metrics(self) -> None:
        self.__sum_acc = 0
        self.__num_samples = 0

    def _module_metrics(self) -> Dict[str, Any]:
        return {"Accuracy": self.__sum_acc / self.__num_samples}

    def forward(self, graph_data, correct_candidate_idxs):
        gnn_output: GnnOutput = self._gnn(**graph_data)

        # Code assumes that there is one slot per-graph, which is true for the original data

        candidate_node_representations = gnn_output.output_node_representations[
            gnn_output.node_idx_references["candidate_nodes"]
        ]  # [num_candidate_nodes, H_out]
        candidate_nodes_slot_idx = gnn_output.node_graph_idx_reference[
            "candidate_nodes"
        ]  # [num_candidate_nodes]

        slot_representations = gnn_output.output_node_representations[
            gnn_output.node_idx_references["slot_node_idx"]
        ]  # [num_slot_nodes, H_out]
        slot_representations_per_candidate = slot_representations[
            candidate_nodes_slot_idx
        ]  # [num_candidate_nodes, H]
        candidate_scores = self.__candidate_scores(
            torch.cat((candidate_node_representations, slot_representations_per_candidate), dim=-1)
        ).squeeze(-1)
        candidate_nodes_logprobs = scatter_log_softmax(
            src=candidate_scores, index=candidate_nodes_slot_idx, eps=0
        )

        with torch.no_grad():
            self.__sum_acc += int(
                (
                    scatter_max(candidate_scores, index=candidate_nodes_slot_idx)[1]
                    == correct_candidate_idxs
                ).sum()
            )
            self.__num_samples += int(slot_representations.shape[0])
        return -candidate_nodes_logprobs[correct_candidate_idxs].mean()


class VarMisuseModel(
    AbstractNeuralModel[VarMisuseSample, TensorizedVarMisuseSample, VarMisuseGraphModel]
):
    def __init__(self, gnn_model: GraphNeuralNetworkModel):
        super().__init__()
        self.__gnn_model = gnn_model

    IDENTIFIER_REGEX: Final = re.compile("[a-zA-Z][a-zA-Z0-9]*")

    @classmethod
    def __add_subtoken_vocab_nodes(cls, graph: GraphData[Tuple[str, bool]]) -> None:
        all_token_nodes = set(chain(*graph.edges["NextToken"]))

        subtoken_edges: List[Tuple[int, int]] = []
        subtoken_node_ids: Dict[str, int] = {}

        for token_node_idx in all_token_nodes:
            token_text = graph.node_information[token_node_idx][0]
            if not cls.IDENTIFIER_REGEX.match(token_text):
                continue
            for subtoken in split_identifier_into_parts(token_text):
                subtoken_node_idx = subtoken_node_ids.get(subtoken)
                if subtoken_node_idx is None:
                    subtoken_node_idx = len(graph.node_information)
                    graph.node_information.append((subtoken, False))
                    subtoken_node_ids[subtoken] = subtoken_node_idx

                subtoken_edges.append((subtoken_node_idx, token_node_idx))

        graph.edges["SubtokenOf"] = subtoken_edges

    def update_metadata_from(self, datapoint: VarMisuseSample) -> None:
        graph = datapoint["ContextGraph"]
        graph_data = GraphData(
            node_information=[
                (graph["NodeLabels"][str(i)], False) for i in range(len(graph["NodeLabels"]))
            ],
            edges=graph["Edges"],
            reference_nodes={},  # This is not needed for metadata loading
        )
        self.__add_subtoken_vocab_nodes(graph_data)
        self.__gnn_model.update_metadata_from(graph_data)

    def build_neural_module(self) -> VarMisuseGraphModel:
        gnn = self.__gnn_model.build_neural_module()
        return VarMisuseGraphModel(gnn)

    def tensorize(self, datapoint: VarMisuseSample) -> Optional[TensorizedVarMisuseSample]:
        graph = datapoint["ContextGraph"]
        all_correct_slots = [
            i
            for i, cand_symbol in enumerate(datapoint["SymbolCandidates"])
            if cand_symbol["IsCorrect"]
        ]
        assert len(all_correct_slots) == 1

        candidate_node_ids = {s["SymbolDummyNode"] for s in datapoint["SymbolCandidates"]}
        graph_data = GraphData(
            node_information=[
                (graph["NodeLabels"][str(i)], i in candidate_node_ids)
                for i in range(len(graph["NodeLabels"]))
            ],
            edges=graph["Edges"],
            reference_nodes={
                "candidate_nodes": [s["SymbolDummyNode"] for s in datapoint["SymbolCandidates"]],
                "slot_node_idx": [datapoint["SlotDummyNode"]],
            },
        )

        if graph_data is None:
            return None

        self.__add_subtoken_vocab_nodes(graph_data)
        tensorized_graph_data = self.__gnn_model.tensorize(graph_data)
        if tensorized_graph_data is None:
            return None

        return TensorizedVarMisuseSample(
            graph=tensorized_graph_data,
            target_candidate_node_idx=all_correct_slots[0],
            num_candidate_nodes=len(datapoint["SymbolCandidates"]),
        )

    def initialize_minibatch(self) -> Dict[str, Any]:
        return {
            "graph_data": self.__gnn_model.initialize_minibatch(),
            "correct_candidate_idxs": [],
            "total_num_candidate_nodes": 0,
        }

    def extend_minibatch_with(
        self, tensorized_datapoint: TensorizedVarMisuseSample, partial_minibatch: Dict[str, Any]
    ) -> bool:
        continue_adding = self.__gnn_model.extend_minibatch_with(
            tensorized_datapoint.graph, partial_minibatch["graph_data"]
        )
        partial_minibatch["correct_candidate_idxs"].append(
            tensorized_datapoint.target_candidate_node_idx
            + partial_minibatch["total_num_candidate_nodes"]
        )
        partial_minibatch["total_num_candidate_nodes"] += tensorized_datapoint.num_candidate_nodes
        return continue_adding

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        return {
            "graph_data": self.__gnn_model.finalize_minibatch(
                accumulated_minibatch_data["graph_data"], device=device
            ),
            "correct_candidate_idxs": torch.tensor(
                accumulated_minibatch_data["correct_candidate_idxs"],
                dtype=torch.int64,
                device=device,
            ),
        }

    def report_accuracy(
        self,
        dataset: Iterator[VarMisuseSample],
        trained_network: VarMisuseGraphModel,
        device: Union[str, torch.device],
    ) -> float:
        trained_network.eval()
        trained_network._reset_module_metrics()
        for mb_data, _ in self.minibatch_iterator(
            self.tensorize_dataset(dataset), device, max_minibatch_size=50
        ):
            with torch.no_grad():
                trained_network(**mb_data)
        return trained_network.report_metrics()["Accuracy"]



from typing_extensions import TypedDict

import numpy as np
import torch
from dpu_utils.mlutils import Vocabulary
from torch import nn
from typing import Any, Counter, Dict, Iterator, List, NamedTuple, Optional, Tuple, Union

from ptgnn.baseneuralmodel import AbstractNeuralModel, ModuleWithMetrics
from ptgnn.baseneuralmodel.utils.data import enforce_not_None
from ptgnn.neuralmodels.gnn import GnnOutput, GraphData, TensorizedGraphData
from ptgnn.neuralmodels.gnn.graphneuralnetwork import GraphNeuralNetwork, GraphNeuralNetworkModel

SuperNodeData = TypedDict(
    "SuperNodeData",
    {
        "name": str,
        "annotation": Optional[str],
    },
    total=False,
)

TypilusGraph = TypedDict(
    "TypilusGraph",
    {
        "nodes": List[str],
        "edges": Dict[str, Dict[str, List[int]]],
        "token-sequence": List[int],
        "supernodes": Dict[str, SuperNodeData],
        "filename": str,
    },
)

Prediction = Tuple[TypilusGraph, Dict[int, Tuple[str, float]]]


class TensorizedGraph2ClassSample(NamedTuple):
    graph: TensorizedGraphData
    supernode_target_classes: List[int]


IGNORED_TYPES = {
    "typing.Any",
    "Any",
    "",
    "typing.NoReturn",
    "NoReturn",
    "nothing",
    "None",
    "T",
    "_T",
    "_T0",
    "_T1",
    "_T2",
    "_T3",
    "_T4",
    "_T5",
    "_T6",
    "_T7",
}


class Graph2ClassModule(ModuleWithMetrics):
    def __init__(self, gnn: GraphNeuralNetwork, num_target_classes: int):
        super().__init__()
        self.__gnn = gnn
        self.__node_to_class = nn.Linear(
            in_features=gnn.output_node_state_dim, out_features=num_target_classes
        )
        nn.init.uniform_(self.__node_to_class.weight)
        nn.init.zeros_(self.__node_to_class.bias)
        self.__loss = nn.CrossEntropyLoss()

    def _reset_module_metrics(self) -> None:
        self.__num_samples = 0
        self.__sum_accuracy = 0

    def _module_metrics(self) -> Dict[str, Any]:
        return {"Accuracy": self.__sum_accuracy / self.__num_samples}

    def _logits(self, graph_mb_data):
        graph_output: GnnOutput = self.__gnn(**graph_mb_data)
        # Gather the output representation of the nodes of interest
        supernode_idxs = graph_output.node_idx_references["supernodes"]
        supernode_graph_idx = graph_output.node_graph_idx_reference["supernodes"]
        supernode_representations = graph_output.output_node_representations[
            supernode_idxs
        ]  # [num_supernodes_in_mb, D]
        return self.__node_to_class(supernode_representations), supernode_graph_idx

    def predict(self, graph_mb_data) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        with torch.no_grad():
            logits, supernode_graph_idx = self._logits(graph_mb_data)
            probs = torch.softmax(logits, dim=-1)
            return torch.max(probs, dim=-1) + (supernode_graph_idx,)

    def forward(self, graph_mb_data, target_classes, original_supernode_idxs):
        logits, _ = self._logits(graph_mb_data)
        with torch.no_grad():
            self.__sum_accuracy += int((torch.argmax(logits, dim=-1) == target_classes).sum())
            self.__num_samples += int(target_classes.shape[0])
        return self.__loss(logits, target_classes)


class Graph2Class(
    AbstractNeuralModel[TypilusGraph, TensorizedGraph2ClassSample, Graph2ClassModule]
):
    def __init__(
        self,
        gnn_model: GraphNeuralNetworkModel,
        max_num_classes: int = 100,
        try_simplify_unks: bool = True,
    ):
        super().__init__()
        self.__gnn_model = gnn_model
        self.max_num_classes = max_num_classes
        self.__try_simplify_unks = try_simplify_unks
        self.__tensorize_samples_with_no_annotation = False
        self.__tensorize_keep_original_supernode_idx = False

    def __convert(self, typilus_graph: TypilusGraph) -> Tuple[GraphData[str, None], List[str]]:
        def get_adj_list(adjacency_dict):
            for from_node_idx, to_node_idxs in adjacency_dict.items():
                from_node_idx = int(from_node_idx)
                for to_idx in to_node_idxs:
                    yield (from_node_idx, to_idx)

        edges = {}
        for edge_type, adj_dict in typilus_graph["edges"].items():
            adj_list: List[Tuple[int, int]] = list(get_adj_list(adj_dict))
            if len(adj_list) > 0:
                edges[edge_type] = np.array(adj_list, dtype=np.int32)
            else:
                edges[edge_type] = np.zeros((0, 2), dtype=np.int32)

        supernode_idxs_with_ground_truth: List[int] = []
        supernode_annotations: List[str] = []
        for supernode_idx, supernode_data in typilus_graph["supernodes"].items():
            if supernode_data["annotation"] in IGNORED_TYPES:
                continue
            if (
                not self.__tensorize_samples_with_no_annotation
                and supernode_data["annotation"] is None
            ):
                continue
            elif supernode_data["annotation"] is None:
                supernode_data["annotation"] = "??"
            supernode_idxs_with_ground_truth.append(int(supernode_idx))
            supernode_annotations.append(enforce_not_None(supernode_data["annotation"]))

        return (
            GraphData[str, None](
                node_information=typilus_graph["nodes"],
                edges=edges,
                reference_nodes={
                    "token-sequence": typilus_graph["token-sequence"],
                    "supernodes": supernode_idxs_with_ground_truth,
                },
            ),
            supernode_annotations,
        )

    # region Metadata Loading
    def initialize_metadata(self) -> None:
        self.__target_class_counter = Counter[str]()

    def update_metadata_from(self, datapoint: TypilusGraph) -> None:
        graph_data, target_classes = self.__convert(datapoint)
        self.__gnn_model.update_metadata_from(graph_data)
        self.__target_class_counter.update(target_classes)

    def finalize_metadata(self) -> None:
        self.__target_vocab = Vocabulary.create_vocabulary(
            self.__target_class_counter,
            max_size=self.max_num_classes + 1,
        )
        del self.__target_class_counter

    # endregion

    def build_neural_module(self) -> Graph2ClassModule:
        return Graph2ClassModule(
            gnn=self.__gnn_model.build_neural_module(), num_target_classes=len(self.__target_vocab)
        )

    def tensorize(self, datapoint: TypilusGraph) -> Optional[TensorizedGraph2ClassSample]:
        graph_data, target_classes = self.__convert(datapoint)
        if len(target_classes) == 0:
            return None  # Sample contains no ground-truth annotations.

        graph_tensorized_data = self.__gnn_model.tensorize(graph_data)

        if graph_tensorized_data is None:
            return None  # Sample rejected by the GNN

        target_class_ids = []
        for target_cls in target_classes:
            if self.__try_simplify_unks and self.__target_vocab.is_unk(target_cls):
                # TODO: Backoff on the type lattice. For now, just erase generics
                generic_start = target_cls.find("[")
                if generic_start != -1:
                    target_cls = target_cls[:generic_start]
            target_class_ids.append(self.__target_vocab.get_id_or_unk(target_cls))

        return TensorizedGraph2ClassSample(
            graph=graph_tensorized_data, supernode_target_classes=target_class_ids
        )

    # region Minibatching
    def initialize_minibatch(self) -> Dict[str, Any]:
        return {
            "graph_mb_data": self.__gnn_model.initialize_minibatch(),
            "target_classes": [],
            "original_supernode_idxs": [],
        }

    def extend_minibatch_with(
        self, tensorized_datapoint: TensorizedGraph2ClassSample, partial_minibatch: Dict[str, Any]
    ) -> bool:
        partial_minibatch["target_classes"].extend(tensorized_datapoint.supernode_target_classes)
        if self.__tensorize_keep_original_supernode_idx:
            partial_minibatch["original_supernode_idxs"].extend(
                tensorized_datapoint.graph.reference_nodes["supernodes"]
            )
        return self.__gnn_model.extend_minibatch_with(
            tensorized_datapoint.graph, partial_minibatch["graph_mb_data"]
        )

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        return {
            "graph_mb_data": self.__gnn_model.finalize_minibatch(
                accumulated_minibatch_data["graph_mb_data"], device
            ),
            "target_classes": torch.tensor(
                accumulated_minibatch_data["target_classes"], dtype=torch.int64, device=device
            ),
            "original_supernode_idxs": accumulated_minibatch_data["original_supernode_idxs"],
        }

    # endregion

    def report_accuracy(
        self,
        dataset: Iterator[TypilusGraph],
        trained_network: Graph2ClassModule,
        device: Union[str, torch.device],
    ) -> float:
        trained_network.eval()
        unk_class_id = self.__target_vocab.get_id_or_unk(self.__target_vocab.get_unk())

        num_correct, num_elements = 0, 0
        for mb_data, _ in self.minibatch_iterator(
            self.tensorize_dataset(dataset), device, max_minibatch_size=50
        ):
            _, predictions, _ = trained_network.predict(mb_data["graph_mb_data"])
            for target_idx, prediction in zip(mb_data["target_classes"], predictions):
                num_elements += 1
                if target_idx == prediction and target_idx != unk_class_id:
                    num_correct += 1
        return num_correct / num_elements

    def predict(
        self,
        data: Iterator[TypilusGraph],
        trained_network: Graph2ClassModule,
        device: Union[str, torch.device],
    ) -> Iterator[Prediction]:
        trained_network.eval()
        with torch.no_grad():
            try:
                self.__tensorize_samples_with_no_annotation = True
                self.__tensorize_keep_original_supernode_idx = True

                for mb_data, original_datapoints in self.minibatch_iterator(
                    self.tensorize_dataset(data, return_input_data=True, parallelize=False),
                    device,
                    max_minibatch_size=50,
                    parallelize=False,
                ):
                    current_graph_idx = 0
                    graph_preds: Dict[int, Tuple[str, float]] = {}

                    probs, predictions, graph_idxs = trained_network.predict(
                        mb_data["graph_mb_data"]
                    )
                    supernode_idxs = mb_data["original_supernode_idxs"]
                    for graph_idx, prediction_prob, prediction_id, supernode_idx in zip(
                        graph_idxs, probs, predictions, supernode_idxs
                    ):
                        if graph_idx != current_graph_idx:
                            yield original_datapoints[current_graph_idx], graph_preds
                            current_graph_idx = graph_idx
                            graph_preds: Dict[int, Tuple[str, float]] = {}

                        predicted_type = self.__target_vocab.get_name_for_id(prediction_id)
                        graph_preds[supernode_idx] = predicted_type, float(prediction_prob)
                    yield original_datapoints[current_graph_idx], graph_preds
            finally:
                self.__tensorize_samples_with_no_annotation = False
                self.__tensorize_keep_original_supernode_idx = False



import numpy as np
from typing import Iterator, List, NamedTuple


class SampleDatapoint(NamedTuple):
    input_features: List[float]
    target_class: bool


class SyntheticData:
    def __init__(self, num_features: int, random_seed: int = None):
        np.random.seed(random_seed)
        self.__num_features = num_features
        self.__weights = np.random.randn(num_features) * 10

    def generate(self, num_points: int) -> Iterator[SampleDatapoint]:
        for _ in range(num_points):
            # Avoid bias so that there is no obvious class imbalance.
            input_features = np.random.randn(self.__num_features) * 5
            yield SampleDatapoint(
                input_features=list(input_features),
                target_class=sum(input_features * self.__weights) >= 0,
            )



from typing_extensions import Final

import logging
import math
import numpy as np
import torch
import torch.nn as nn
from dpu_utils.mlutils import Vocabulary
from torch_scatter import scatter_add
from torch_scatter.composite import scatter_log_softmax, scatter_logsumexp
from typing import Any, Counter, Dict, List, NamedTuple, Tuple, Union

from ptgnn.baseneuralmodel import AbstractNeuralModel


class DecoderData(NamedTuple):
    input_elements: List[str]
    target_data: List[str]


class TokenizedOutput(NamedTuple):
    token_ids: List[int]
    length: int
    num_input_elements: int
    # The indices that can be copied from the input elements at each point, if any.
    copyable_elements: List[np.ndarray]


class GruCopyingDecoder(nn.Module):
    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        hidden_size: int,
        memories_hidden_dim: int,
        unk_id: int,
        dropout_rate: float,
    ):
        super().__init__()
        self.__embedding_layer = nn.Embedding(
            num_embeddings=vocabulary_size, embedding_dim=embedding_size
        )
        self.__output_gru = nn.GRU(
            input_size=embedding_size, hidden_size=hidden_size, num_layers=1, batch_first=True
        )

        self.__unk_id = unk_id

        self.__memories_to_standard_attention = nn.Linear(
            in_features=memories_hidden_dim, out_features=hidden_size, bias=False
        )
        self.__memories_to_copy_attention = nn.Linear(
            in_features=memories_hidden_dim, out_features=hidden_size, bias=False
        )
        self.__hidden_to_vocab = nn.Parameter(0.01 * torch.randn((2 * hidden_size, embedding_size)))
        self.__vocab_bias = nn.Parameter(torch.zeros(vocabulary_size))
        self.__dropout = nn.Dropout(dropout_rate)

    def _compute_logprobs(
        self, initial_states, input_memories, input_memories_origin_idx, input_token_ids
    ):
        """
        :param input_memories: [num-inputs-flattened, D]
        :param input_memories_origin_idx: [num-inputs-flattened]
        :param initial_states: [num-targets, H]
        :param input_token_ids: [num-targets, max-seq-size-1]

        :return: the logprobs for all copying locations and the logprobs for all elements in the vocabulary
        """
        target_token_embeddings = self.__dropout(
            self.__embedding_layer(input_token_ids)
        )  # [num-targets, max-seq-size - 1, H]
        output_states, output_gru_state = self.__output_gru(
            target_token_embeddings, initial_states.unsqueeze(0)
        )
        output_states = output_states.contiguous()  # [num-targets, max-seq-size, H]

        # Standard and copy attention representations
        standard_attention_reps = self.__memories_to_standard_attention(
            input_memories
        )  # [num-inputs-flattened, H]

        copy_attention_reps = self.__memories_to_copy_attention(
            input_memories
        )  # [num-inputs-flattened, H]
        copy_attention_reps = self.__dropout(copy_attention_reps)

        # Compute attention scores [num-inputs-flattened, max-seq-size]
        output_states_per_input = output_states[
            input_memories_origin_idx
        ]  # [num-inputs-flattened, max-seq-size, H]
        standard_attention_scores = torch.einsum(
            "ilh,ih->il", output_states_per_input, standard_attention_reps
        )  # [num-inputs-flattened, max-seq-size]
        copy_attention_scores = torch.einsum(
            "ilh,ih->il", output_states_per_input, copy_attention_reps
        )  # [num-inputs-flattened, max-seq-size]

        # For some reason scatter_softmax doesn't work here
        standard_attention_logprobs = scatter_log_softmax(
            standard_attention_scores, index=input_memories_origin_idx, dim=0, eps=0
        )  # [num-inputs-flattened, max-seq-size]

        # Compute standard attention output
        standard_attention_mul = torch.einsum(
            "il,ih->ilh", torch.exp(standard_attention_logprobs), standard_attention_reps
        )
        standard_attention_out = scatter_add(
            standard_attention_mul, index=input_memories_origin_idx, dim=0
        )  # [num-targets, max-seq-size, H]
        target_scores = (
            torch.einsum(
                "blh,hd,vd->blv",
                torch.cat((self.__dropout(standard_attention_out), output_states), dim=-1),
                self.__hidden_to_vocab,
                self.__dropout(self.__embedding_layer.weight),
            )
            + self.__vocab_bias
        )  # [num-targets, max-seq-size, vocab-size]

        # A "manual" log_softmax
        total_copy_scores = scatter_logsumexp(
            copy_attention_scores, index=input_memories_origin_idx, dim=0, eps=0
        )  # [num-targets, max-seq-size]
        all_scores = torch.cat(
            (target_scores, total_copy_scores.unsqueeze(-1)), dim=-1
        )  # [num-targets, max-seq-size, vocab-size + 1]
        normalizing_const = torch.logsumexp(all_scores, dim=-1)  # [num-targets, max-seq-size]

        target_logprobs = target_scores - normalizing_const.unsqueeze(
            -1
        )  # [num-targets, max-seq-size, vocab-size]
        copy_logprobs = (
            copy_attention_scores - normalizing_const[input_memories_origin_idx]
        )  # [num-inputs-flattened, max-seq-size]

        # Do probabilities actually sum up to 1?
        # sum_logcopy = scatter_logsumexp(copy_logprobs, index=input_memories_origin_idx, dim=0, eps=0) # [num-targets, max-seq-size]
        # sum_targets = torch.logsumexp(target_logprobs, dim=-1)                          # [num-targets, max-seq-size]
        # torch.allclose(torch.logsumexp(torch.stack([sum_logcopy, sum_targets], dim=-1), dim=-1), torch.zeros_like(sum_logcopy), atol=.0001)

        return copy_logprobs, target_logprobs, output_gru_state

    def forward(
        self,
        *,
        input_memories,
        input_memories_origin_idx,
        initial_states,
        target_token_ids,
        copyable_elements_idxs,
        copyable_elements_sample_idxs,
        target_lengths
    ):
        """
        :param input_memories: [num-inputs-flattened, D]
        :param input_memories_origin_idx: [num-inputs-flattened]
        :param initial_states: [num-targets, H]
        :param target_token_ids: [num-targets, max-seq-size]
        :param target_lengths: [num-targets]
        :param copyable_elements_idxs: [num-copyable-elements]
        :param copyable_elements_sample_idxs: [num-copyable-elements]

        :return: the loss function.
        """
        copy_logprobs, target_logprobs, _ = self._compute_logprobs(
            initial_states, input_memories, input_memories_origin_idx, target_token_ids[:, :-1]
        )

        # Get loss. UNKs are only predicted if we cannot copy.
        num_valid_copy_actions = scatter_add(
            src=torch.ones_like(copyable_elements_sample_idxs),
            index=copyable_elements_sample_idxs,
            dim_size=target_token_ids.shape[0] * (target_token_ids.shape[1] - 1),
        )
        locations_with_valid_copy_actions = (
            num_valid_copy_actions.reshape(target_token_ids.shape[0], target_token_ids.shape[1] - 1)
            > 0
        )
        unk_prediction_locations = target_token_ids[:, 1:] == self.__unk_id
        mask = locations_with_valid_copy_actions & unk_prediction_locations

        correct_generation_logprobs = torch.gather(
            target_logprobs, index=target_token_ids[:, 1:].unsqueeze(-1), dim=-1
        ).squeeze(
            -1
        )  # [num-targets, max-seq-size -1]
        correct_generation_logprobs.masked_fill_(mask, -math.inf)

        correct_copy_logprobs = scatter_logsumexp(
            src=copy_logprobs.flatten()[copyable_elements_idxs],
            index=copyable_elements_sample_idxs,
            dim=0,
            dim_size=target_token_ids.shape[0] * (target_token_ids.shape[1] - 1),
            eps=0,
        )  # [num-targets * (max-seq-size-1)]
        correct_copy_logprobs = correct_copy_logprobs.view(
            target_token_ids.shape[0], target_token_ids.shape[1] - 1
        )

        any_correct_action_logprob = torch.logsumexp(
            torch.stack((correct_generation_logprobs, correct_copy_logprobs)), dim=0
        )  # [num-targets, max-seq-size]

        mask = torch.arange(
            any_correct_action_logprob.shape[1], device=target_lengths.device
        ).unsqueeze(0) < target_lengths.unsqueeze(1)
        per_seq_loss = (any_correct_action_logprob * mask.float()).sum(dim=-1) / mask.float().sum(
            dim=-1
        )

        return -per_seq_loss.mean()


class GruCopyingDecoderModel(AbstractNeuralModel[DecoderData, TokenizedOutput, GruCopyingDecoder]):
    """A GRU copying decoder that accepts an variable-sized set of encodings."""

    LOGGER: Final = logging.getLogger(__name__)

    def __init__(
        self,
        *,
        max_seq_len: int = 8,
        hidden_size=128,
        embedding_size: int = 256,
        memories_hidden_dim: int = 128,
        vocabulary_max_size: int = 20000,
        vocabulary_count_threshold: int = 5,
        dropout_rate: float = 0.2
    ):
        super().__init__()
        self.max_seq_len: Final = max_seq_len
        self.hidden_size: Final = hidden_size
        self.embedding_size: Final = embedding_size
        self.memories_hidden_dim: Final = memories_hidden_dim
        self.vocabulary_max_size: Final = vocabulary_max_size
        self.vocabulary_count_threshold: Final = vocabulary_count_threshold
        self.dropout_rate: Final = dropout_rate

    @property
    def END(self) -> str:
        return "%END%"

    @property
    def START(self) -> str:
        return "%START%"

    # region Metadata
    def initialize_metadata(self) -> None:
        self.__token_counter = Counter[str]()

    def update_metadata_from(self, datapoint: DecoderData) -> None:
        self.__token_counter.update(datapoint.target_data)

    def finalize_metadata(self) -> None:
        self.__token_counter[self.START] = 1000000
        self.__token_counter[self.END] = 1000000
        self.__output_vocabulary = Vocabulary.create_vocabulary(
            self.__token_counter,
            max_size=self.vocabulary_max_size,
            count_threshold=self.vocabulary_count_threshold,
        )
        self.LOGGER.info("Output vocabulary Size %s", len(self.__output_vocabulary))
        del self.__token_counter

    def build_neural_module(self) -> GruCopyingDecoder:
        return GruCopyingDecoder(
            vocabulary_size=len(self.__output_vocabulary),
            embedding_size=self.embedding_size,
            hidden_size=self.hidden_size,
            memories_hidden_dim=self.memories_hidden_dim,
            unk_id=self.__output_vocabulary.get_id_or_unk(self.__output_vocabulary.get_unk()),
            dropout_rate=self.dropout_rate,
        )

    # endregion

    def tensorize(self, datapoint: DecoderData) -> TokenizedOutput:
        max_seq_len = self.max_seq_len
        target_with_start_end = [self.START] + datapoint.target_data + [self.END]
        target_with_start_end = target_with_start_end[:max_seq_len]

        seq_len = min(len(target_with_start_end), max_seq_len)

        return TokenizedOutput(
            token_ids=self.__output_vocabulary.get_id_or_unk_multiple(target_with_start_end),
            length=seq_len,
            num_input_elements=len(datapoint.input_elements),
            # Which elements can be copied at this step?
            copyable_elements=[
                np.array(
                    [
                        i
                        for i, input_element in enumerate(datapoint.input_elements)
                        if input_element == target_token
                    ],
                    dtype=np.int32,
                )
                for target_token in target_with_start_end[1:]
            ],
        )

    def initialize_minibatch(self) -> Dict[str, Any]:
        return {
            "target_token_ids": [],
            "target_seq_lengths": [],
            # The indices of the input elements that can be copied for each timestep in range(max-seq-len-1)
            "copyable_elements_idxs": [],
            "num_input_elements": [],
        }

    def extend_minibatch_with(
        self, tensorized_datapoint: TokenizedOutput, partial_minibatch: Dict[str, Any]
    ) -> bool:
        partial_minibatch["target_token_ids"].append(tensorized_datapoint.token_ids)
        partial_minibatch["target_seq_lengths"].append(tensorized_datapoint.length)
        partial_minibatch["copyable_elements_idxs"].append(tensorized_datapoint.copyable_elements)
        partial_minibatch["num_input_elements"].append(tensorized_datapoint.num_input_elements)
        return True

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        max_seq_length = max(accumulated_minibatch_data["target_seq_lengths"])
        num_elements = len(accumulated_minibatch_data["target_token_ids"])

        target_token_ids = np.zeros((num_elements, max_seq_length), dtype=np.int32)

        # These need to index into index into [num-inputs-flattened, max-seq-size - 1] memories
        copyable_elements_idxs: List[int] = []
        # Each group of elements is one entry of a [num-target, max-seq-size - 1]
        copyable_elements_sample_idxs: List[int] = []

        # Count how many input sequences have been seen up to each point of the iteration
        offset_num_input_elements_so_far = 0
        for sample_idx, (token_ids, copyable_elements, num_input_elements_for_sample) in enumerate(
            zip(
                accumulated_minibatch_data["target_token_ids"],
                accumulated_minibatch_data["copyable_elements_idxs"],
                accumulated_minibatch_data["num_input_elements"],
            )
        ):
            target_token_ids[sample_idx, : len(token_ids)] = token_ids

            sample_row_starting_idx = sample_idx * (max_seq_length - 1)
            for timestep_idx, copyable_elements_at_timestep in enumerate(copyable_elements):
                flattened_index = (
                    sample_row_starting_idx + timestep_idx
                )  # The -1 is here since the flattened output does not have a prediction for the last element
                # Confusingly this is indexed by timestep. So (after the offset) we have a flattened out [num-inputs-flattened, max-seq-size-1]
                copyable_elements_idxs.extend(
                    offset_num_input_elements_so_far
                    + copyable_elements_at_timestep * (max_seq_length - 1)
                    + timestep_idx
                )
                copyable_elements_sample_idxs.extend(
                    flattened_index for _ in range(len(copyable_elements_at_timestep))
                )
                # Add one full sequence of inputs here
            offset_num_input_elements_so_far += num_input_elements_for_sample * (max_seq_length - 1)

        return {
            "target_token_ids": torch.tensor(target_token_ids, dtype=torch.int64, device=device),
            "copyable_elements_idxs": torch.tensor(
                copyable_elements_idxs, dtype=torch.int64, device=device
            ),
            "copyable_elements_sample_idxs": torch.tensor(
                copyable_elements_sample_idxs, dtype=torch.int64, device=device
            ),
            "target_lengths": torch.tensor(
                accumulated_minibatch_data["target_seq_lengths"], dtype=torch.int64, device=device
            ),
        }

    def greedy_decode(
        self,
        *,
        input_concrete_values: List[str],
        input_memories,
        input_memories_origin_idx,
        initial_states,
        neural_model: GruCopyingDecoder
    ) -> List[Tuple[List[str], float]]:
        output_vocab = self.__output_vocabulary
        batch_size = initial_states.shape[0]
        assert len(input_concrete_values) == input_memories.shape[0]

        current_decoder_states = initial_states  # [num-targets, H]
        next_token_sequences = torch.tensor(
            [[output_vocab.get_id_or_unk(self.START)]] * batch_size, device=input_memories.device
        )  # [num-targets x 1]

        predicted_tokens: List[List[str]] = [[] for _ in range(batch_size)]
        predicted_logprobs: List[float] = [0.0 for _ in range(batch_size)]
        sample_id_done = np.zeros(batch_size, dtype=np.bool)

        for i in range(self.max_seq_len):
            copy_logprobs, target_logprobs, output_decoder_state = neural_model._compute_logprobs(
                current_decoder_states,
                input_memories,
                input_memories_origin_idx,
                next_token_sequences,
            )
            current_decoder_states = output_decoder_state.squeeze(0)  # [num-targets, H]

            # To speed things up, look up the top 100 vocab tokens
            topk_vocab_logprobs, topk_vocab_idxs = torch.topk(
                target_logprobs.squeeze(1), min(100, target_logprobs.shape[-1]), dim=-1
            )  # [num-targets, 100], [num-targets, 100]

            # Move to CPU and show as numpy
            topk_vocab_logprobs = topk_vocab_logprobs.cpu().numpy()
            topk_vocab_idxs = topk_vocab_idxs.cpu().numpy()
            copy_logprobs = copy_logprobs.squeeze(1).cpu().numpy()

            target_token_predictions = [
                {
                    output_vocab.get_name_for_id(int(token_idx)): token_logprob
                    for token_idx, token_logprob in zip(
                        topk_vocab_idxs[batch_idx], topk_vocab_logprobs[batch_idx]
                    )
                }
                for batch_idx in range(batch_size)
            ]  # for each element in the batch, the logprobs of the target predictions

            for batch_idx, concrete_value, copy_logprob in zip(
                input_memories_origin_idx.cpu().numpy(), input_concrete_values, copy_logprobs
            ):
                predictions = target_token_predictions[batch_idx]
                predictions[concrete_value] = np.logaddexp(
                    predictions.get(concrete_value, -math.inf), copy_logprob
                )

            predicted_tokens_for_this_step = []
            for batch_idx, predictions in enumerate(target_token_predictions):
                if sample_id_done[batch_idx]:
                    predicted_tokens_for_this_step.append(self.END)
                    continue

                predicted_token, predicted_logprob = max(
                    target_token_predictions[batch_idx].items(), key=lambda x: x[1]
                )

                if predicted_token == self.END:
                    sample_id_done[batch_idx] = True
                else:
                    predicted_tokens[batch_idx].append(predicted_token)

                predicted_tokens_for_this_step.append(predicted_token)
                predicted_logprobs[batch_idx] += predicted_logprob

            next_token_sequences = torch.tensor(
                [[output_vocab.get_id_or_unk(t)] for t in predicted_tokens_for_this_step],
                device=input_memories.device,
            )

        return list(zip(predicted_tokens, predicted_logprobs))



from .abstractmessagepassing import AbstractMessagePassingLayer
from .gatedmessagepassing import GatedMessagePassingLayer
from .globalgraphexchange import AbstractGlobalGraphExchange, GruGlobalStateUpdate
from .mlpmessagepassing import MlpMessagePassingLayer
from .pna_aggregation import PnaMessageAggregation
from .residuallayers import LinearResidualLayer, MeanResidualLayer



import numpy as np
import torch
from torch import nn
from typing import Any, Dict, Iterable, NamedTuple, Optional, Union

from ptgnn.baseneuralmodel import AbstractNeuralModel, ModuleWithMetrics
from ptgnn.baseneuralmodel.utils.data import enforce_not_None
from ptgnn.implementations.ppi.dataloader import PPIGraphSample
from ptgnn.neuralmodels.gnn.graphneuralnetwork import GraphNeuralNetwork, GraphNeuralNetworkModel
from ptgnn.neuralmodels.gnn.structs import GnnOutput, GraphData, TensorizedGraphData


class PPIClassification(ModuleWithMetrics):
    def __init__(self, gnn: GraphNeuralNetwork, num_target_classes: int):
        super().__init__()
        self.__gnn = gnn
        self.__output_representation_to_logits = nn.Linear(
            in_features=gnn.output_node_state_dim, out_features=num_target_classes
        )
        torch.nn.init.xavier_uniform(self.__output_representation_to_logits.weight)
        torch.nn.init.zeros_(self.__output_representation_to_logits.bias)
        self.__output_representation_to_logits

    def _reset_module_metrics(self) -> None:
        self.__num_samples = 0
        self.__sum_f1 = 0.0
        self.__sum_pr = 0.0
        self.__sum_re = 0.0

    def _module_metrics(self) -> Dict[str, Any]:
        return {
            "f1_score": self.__sum_f1 / self.__num_samples,
            "pr_score": self.__sum_pr / self.__num_samples,
            "re_score": self.__sum_re / self.__num_samples,
        }

    def forward(self, graph_data, targets):
        gnn_output: GnnOutput = self.__gnn(**graph_data)
        target_logits = self.__output_representation_to_logits(
            gnn_output.output_node_representations
        )

        with torch.no_grad():
            predictions = torch.sigmoid(target_logits) >= 0.5

            true_positives = (predictions & targets).sum().float()
            false_positives = (predictions & (~targets)).sum().float()
            false_negatives = ((~predictions) & targets).sum().float()

            precision = true_positives / (true_positives + false_positives + 1e-10)
            recall = true_positives / (true_positives + false_negatives + 1e-10)
            fscore = 2 * precision * recall / (precision + recall + 1e-10)
            num_samples = int(predictions.shape[0])
            self.__sum_f1 += float(fscore.cpu()) * num_samples
            self.__sum_pr += float(precision.cpu()) * num_samples
            self.__sum_re += float(recall.cpu()) * num_samples
            self.__num_samples += num_samples

        losses = nn.functional.binary_cross_entropy_with_logits(
            input=target_logits, target=targets.float(), reduction="none"
        )
        return losses.sum(dim=-1).mean()


class TensorizedPPIData(NamedTuple):
    gnn_data: TensorizedGraphData[np.ndarray]
    targets: np.ndarray


class PPIMulticlassClassification(
    AbstractNeuralModel[PPIGraphSample, TensorizedPPIData, PPIClassification]
):
    def __init__(self, gnn_model: GraphNeuralNetworkModel[np.ndarray, np.ndarray]):
        super().__init__()
        self.__gnn_model = gnn_model

    def initialize_metadata(self) -> None:
        self.__num_target_labels: Optional[int] = None

    def update_metadata_from(self, datapoint: PPIGraphSample) -> None:
        self.__gnn_model.update_metadata_from(
            GraphData(
                node_information=datapoint.node_features,
                edges={f"e{i}": a for i, a in enumerate(datapoint.adjacency_lists)},
                reference_nodes={},
            ),
        )
        if self.__num_target_labels is None:
            self.__num_target_labels = datapoint.node_labels.shape[1]
        else:
            assert self.__num_target_labels == datapoint.node_labels.shape[1]

    def build_neural_module(self) -> PPIClassification:
        gnn = self.__gnn_model.build_neural_module()
        return PPIClassification(gnn, enforce_not_None(self.__num_target_labels))

    def tensorize(self, datapoint: PPIGraphSample) -> Optional[TensorizedPPIData]:
        graph_tensors = self.__gnn_model.tensorize(
            GraphData[np.ndarray](
                node_information=datapoint.node_features,
                edges={f"e{i}": a for i, a in enumerate(datapoint.adjacency_lists)},
                reference_nodes={},
            )
        )

        if graph_tensors is None:
            return None

        return TensorizedPPIData(graph_tensors, datapoint.node_labels)

    def initialize_minibatch(self) -> Dict[str, Any]:
        return {"graph_data": self.__gnn_model.initialize_minibatch(), "labels": []}

    def extend_minibatch_with(
        self, tensorized_datapoint: TensorizedPPIData, partial_minibatch: Dict[str, Any]
    ) -> bool:
        continue_adding = self.__gnn_model.extend_minibatch_with(
            tensorized_datapoint.gnn_data, partial_minibatch["graph_data"]
        )
        partial_minibatch["labels"].append(tensorized_datapoint.targets)
        return continue_adding

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        return {
            "graph_data": self.__gnn_model.finalize_minibatch(
                accumulated_minibatch_data["graph_data"], device
            ),
            "targets": torch.tensor(
                np.concatenate(accumulated_minibatch_data["labels"], axis=0),
                dtype=torch.bool,
                device=device,
            ),
        }

    def report_metrics(
        self,
        dataset: Iterable[PPIGraphSample],
        trained_network: PPIClassification,
        device: Union[str, torch.device],
    ) -> Dict[str, float]:
        trained_network.eval()
        trained_network._reset_module_metrics()
        for mb_data, _ in self.minibatch_iterator(
            self.tensorize_dataset(iter(dataset)), device, max_minibatch_size=50
        ):
            with torch.no_grad():
                trained_network(**mb_data)
        return trained_network.report_metrics()



from .varsizedsummary import (
    AbstractVarSizedElementReduce,
    ElementsToSummaryRepresentationInput,
    MultiheadSelfAttentionVarSizedElementReduce,
    NormalizedWeightsVarSizedElementReduce,
    SelfAttentionVarSizedElementReduce,
    SimpleVarSizedElementReduce,
    WeightedSumVarSizedElementReduce,
)



#!/usr/bin/env python
"""
Usage:
    test_model.py [options] MODEL_FILENAME TEST_DATA_PATH

Options:
    --azure-info=<path>        Azure authentication information file (JSON). Used to load data from Azure storage.
    --max-num-epochs=<epochs>  The maximum number of epochs to run training for. [default: 100]
    --minibatch-size=<size>    The minibatch size. [default: 100]
    --restore-path=<path>      The path to previous model file for starting from previous checkpoint.
    --quiet                    Do not show progress bar.
    -h --help                  Show this screen.
    --debug                    Enable debug routines. [default: False]
"""
import logging
import numpy as np
from docopt import docopt
from dpu_utils.utils import RichPath, run_and_debug
from jellyfish import jaro_winkler
from pathlib import Path
from typing import Tuple

from ptgnn.implementations.graph2seq.graph2seq import Graph2Seq, Graph2SeqModule


def run(arguments):
    azure_info_path = arguments.get("--azure-info", None)

    data_path = RichPath.create(arguments["TEST_DATA_PATH"], azure_info_path)
    data = list(data_path.read_as_jsonl())

    model_path = Path(arguments["MODEL_FILENAME"])
    model, nn = Graph2Seq.restore_model(
        model_path, device="cpu"
    )  # type: Tuple[Graph2Seq, Graph2SeqModule]
    nn.reset_metrics()

    all_responses = model.greedy_decode(data, nn, device="cpu")
    correct_elements, jw_sim, num_elements = 0, 0, 0
    sum_f1, sum_precision, sum_recall = 0.0, 0.0, 0.0
    for (res_tokens, res_logprob), actual_data in zip(all_responses, data):
        num_elements += 1
        print(f'{actual_data["method_name"]} -> {res_tokens} ({np.exp(res_logprob):.2f})')

        jw_sim += jaro_winkler("".join(actual_data["method_name"]), "".join(res_tokens))
        if actual_data["method_name"] == res_tokens:
            correct_elements += 1
        res_tokens = set(res_tokens)
        res_tokens.discard("%UNK%")
        ground_tokens = set(actual_data["method_name"])
        if len(res_tokens) > 0:
            precision = len(res_tokens & ground_tokens) / len(res_tokens)
        else:
            precision = 0
        recall = len(res_tokens & ground_tokens) / len(ground_tokens)

        if precision + recall > 0:
            sum_f1 += 2 * recall * precision / (precision + recall)
            sum_precision += precision
            sum_recall += recall

    print(f"Acc {correct_elements / num_elements: %}  ({correct_elements}/{num_elements})")
    print(f"F1 {sum_f1 / num_elements}")
    print(f"Pr {sum_precision / num_elements}  Re {sum_recall / num_elements}")
    print(f"JW Sim {jw_sim / num_elements}")


if __name__ == "__main__":
    args = docopt(__doc__)
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)-15s %(name)-5s %(levelname)-8s %(message)s"
    )
    run_and_debug(lambda: run(args), args.get("--debug", False))



#!/usr/bin/env python
"""
Usage:
    trainandtest.py [options] TRAIN_DATA_PATH VALID_DATA_PATH TEST_DATA_PATH MODEL_FILENAME

Options:
    --aml                      Run this in Azure ML
    --amp                      Enable automatic mixed precision.
    --azure-info=<path>        Azure authentication information file (JSON). Used to load data from Azure storage.
    --max-num-epochs=<epochs>  The maximum number of epochs to run training for. [default: 100]
    --minibatch-size=<size>    The minibatch size. [default: 300]
    --restore-path=<path>      The path to previous model file for starting from previous checkpoint.
    --sequential-run           Do not parallelize data loading. Makes debugging easier.
    --quiet                    Do not show progress bar.
    -h --help                  Show this screen.
    --debug                    Enable debug routines. [default: False]
"""

from docopt import docopt
from dpu_utils.utils import run_and_debug

from ptgnn.implementations.graph2seq import test, train

if __name__ == "__main__":
    args = docopt(__doc__)
    run_and_debug(lambda: train.run(args), args.get("--debug", False))
    run_and_debug(lambda: test.run(args), args.get("--debug", False))



from typing_extensions import Final, Protocol

import json
import logging
import math
import time
import torch
from pathlib import Path
from tqdm import tqdm
from typing import Callable, Dict, Generic, Iterable, List, Optional, TypeVar

from ptgnn.baseneuralmodel.abstractneuralmodel import AbstractNeuralModel
from ptgnn.baseneuralmodel.modulewithmetrics import ModuleWithMetrics
from ptgnn.baseneuralmodel.utils.data import MemorizedDataIterable

TRawDatapoint = TypeVar("TRawDatapoint")
TTensorizedDatapoint = TypeVar("TTensorizedDatapoint")
TNeuralModule = TypeVar("TNeuralModule", bound=ModuleWithMetrics)
ModelType = AbstractNeuralModel[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]
EndOfEpochHook = Callable[[ModelType, TNeuralModule, int, Dict], None]

__all__ = ["ModelTrainer", "AbstractScheduler", "EndOfEpochHook"]


class AbstractScheduler(Protocol):
    def step(self, epoch_idx: int, epoch_step: int) -> None:
        ...


class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
    """
    A trainer for `AbstractComponent`s. Used mainly for supervised learning.

    Create a `ComponentTrainer` by passing a `AbstractComponent` in the constructor.
    Invoke `train()` to initiate the training loop. The root `TNeuralModule` should return a scalar loss.
    """

    LOGGER: Final = logging.getLogger(__name__)

    def __init__(
        self,
        model: AbstractNeuralModel[TRawDatapoint, TTensorizedDatapoint, TNeuralModule],
        checkpoint_location: Path,
        *,
        max_num_epochs: int = 100,
        minibatch_size: int = 200,
        optimizer_creator: Optional[
            Callable[[Iterable[torch.Tensor]], torch.optim.Optimizer]
        ] = None,
        scheduler_creator: Optional[Callable[[torch.optim.Optimizer], AbstractScheduler]] = None,
        clip_gradient_norm: Optional[float] = None,
        target_validation_metric: Optional[str] = None,
        target_validation_metric_higher_is_better: bool = False,
        enable_amp: bool = False,
    ):
        """
        :param model: The Component to be built and trained
        :param checkpoint_location: The location where the trained model will be checkpointed and saved.
        :param max_num_epochs: The maximum number of epochs to run training for.
        :param minibatch_size: The maximum size of the minibatch (`BaseComponent`s can override this
            by detecting full minibatches and returning False in `extend_minibatch_by_sample`)
        :param optimizer_creator: An optional function that accepts an iterable of the training parameters
            (pyTorch tensors) and returns a PyTorch optimizer.
        :param scheduler_creator: An optional function that accepts an optimizer and creates a scheduler
            implementing `AbstractScheduler`. This could be a wrapper for existing learning schedulers.
            The scheduler will be invoked at after each training step.
        """
        self.__model = model
        self.__neural_network: Optional[TNeuralModule] = None
        self.__checkpoint_location = checkpoint_location

        self.__max_num_epochs = max_num_epochs
        self.__minibatch_size = minibatch_size
        if optimizer_creator is None:
            self.__create_optimizer = lambda p: torch.optim.Adam(p)
        else:
            self.__create_optimizer = optimizer_creator

        self.__create_scheduler = scheduler_creator

        self.__metadata_finalized_hooks: List[Callable[[ModelType], None]] = []
        self.__training_start_hooks: List[
            Callable[[ModelType, TNeuralModule, torch.optim.Optimizer], None]
        ] = []
        self.__train_epoch_end_hooks: List[EndOfEpochHook] = []
        self.__validation_epoch_end_hooks: List[EndOfEpochHook] = []
        self.__clip_gradient_norm = clip_gradient_norm
        self.__enable_amp = enable_amp

        self.__target_metric = target_validation_metric
        if target_validation_metric is not None:
            self.__target_metric_higher_is_better = target_validation_metric_higher_is_better
        else:
            assert (
                not target_validation_metric_higher_is_better
            ), "When no explicit metric is passed, the validation loss will be used."
            self.__target_metric_higher_is_better = False

    @property
    def model(self) -> AbstractNeuralModel[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]:
        return self.__model

    @property
    def neural_module(self) -> TNeuralModule:
        if self.__neural_network is None:
            raise Exception("Neural Network Module has not been built.")
        return self.__neural_network

    @neural_module.setter
    def neural_module(self, nn: TNeuralModule):
        self.__neural_network = nn

    def load_metadata_and_create_network(
        self, training_data: Iterable[TRawDatapoint], parallelize: bool, show_progress_bar: bool
    ) -> None:
        return self.__load_metadata_and_create_network(
            training_data, parallelize, show_progress_bar
        )

    def __load_metadata_and_create_network(
        self, training_data: Iterable[TRawDatapoint], parallelize: bool, show_progress_bar: bool
    ) -> None:
        """
        Compute model metadata by doing a full pass over the training data.
        """
        self.__model.compute_metadata(
            iter(
                tqdm(
                    training_data,
                    desc="Loading Metadata",
                    leave=False,
                    disable=not show_progress_bar,
                )
            ),
            parallelize,
        )
        self.__neural_network = self.__model.build_neural_module()
        self.LOGGER.info(
            "Model metadata loaded. The following model was created:\n %s", self.__neural_network
        )

        for m in self.__metadata_finalized_hooks:
            m(self.__model)

        self.LOGGER.info(
            "Model Definition:\n %s",
            json.dumps(dict(self.__model.model_definition), indent=2),
        )

        self.LOGGER.info("Saving model with finalized metadata.")
        self.__save_checkpoint()

    def __save_checkpoint(self) -> None:
        self.__model.save(self.__checkpoint_location, self.neural_module)

    def __restore_checkpoint(self, device=None) -> None:
        _, self.neural_module = self.__model.restore_model(
            self.__checkpoint_location, device=device
        )

    def register_model_metadata_finalized_hook(self, hook: Callable[[ModelType], None]) -> None:
        self.__metadata_finalized_hooks.append(hook)

    def register_training_start_hook(
        self, hook: Callable[[ModelType, TNeuralModule, torch.optim.Optimizer], None]
    ) -> None:
        self.__training_start_hooks.append(hook)

    def register_train_epoch_end_hook(self, hook: EndOfEpochHook) -> None:
        self.__train_epoch_end_hooks.append(hook)

    def register_validation_epoch_end_hook(self, hook: EndOfEpochHook) -> None:
        self.__validation_epoch_end_hooks.append(hook)

    def _run_training(
        self,
        training_tensors,
        epoch,
        device,
        exponential_running_average_factor,
        optimizer,
        parallelize,
        scheduler,
        show_progress_bar,
        shuffle_input: bool = True,
    ):
        sum_epoch_loss, running_avg_loss, num_minibatches, num_samples = 0.0, 0.0, 0, 0
        start_time = time.time()
        self.neural_module.train()

        scaler = torch.cuda.amp.GradScaler(enabled=self.__enable_amp)
        with tqdm(desc="Training", disable=not show_progress_bar, leave=False) as progress_bar:
            for step_idx, (mb_data, raw_samples) in enumerate(
                self.__model.minibatch_iterator(
                    training_tensors(),
                    device=device,
                    max_minibatch_size=self.__minibatch_size,
                    yield_partial_minibatches=False,
                    shuffle_input=shuffle_input,
                    parallelize=parallelize,
                )
            ):
                optimizer.zero_grad()
                with torch.cuda.amp.autocast(enabled=self.__enable_amp):
                    mb_loss = self.neural_module(**mb_data)
                    if torch.isnan(mb_loss):
                        raise Exception("Loss has a NaN value.")

                    scaler.scale(mb_loss).backward()

                    if self.__clip_gradient_norm is not None:
                        scaler.unscale_(optimizer)
                        torch.nn.utils.clip_grad_norm_(
                            self.neural_module.parameters(recurse=True), self.__clip_gradient_norm
                        )

                    scaler.step(optimizer)
                    scaler.update()
                    if scheduler is not None:
                        scheduler.step(epoch_idx=epoch, epoch_step=step_idx)

                num_minibatches += 1
                num_samples += len(raw_samples)
                mb_loss = float(mb_loss.cpu())
                sum_epoch_loss += mb_loss
                if num_minibatches == 1:  # First minibatch
                    running_avg_loss = mb_loss
                else:
                    running_avg_loss = (
                        exponential_running_average_factor * running_avg_loss
                        + (1 - exponential_running_average_factor) * mb_loss
                    )
                progress_bar.update()
                progress_bar.set_postfix(Loss=f"{running_avg_loss:.2f}")

        elapsed_time = time.time() - start_time
        self.LOGGER.info(
            "Training complete in %.1fsec [%.2f samples/sec]",
            elapsed_time,
            num_samples / elapsed_time,
        )
        assert (
            num_minibatches > 0
        ), "No training minibatches were created. The minibatch size may be too large or the training dataset size too small."
        self.LOGGER.info("Epoch %i: Train Loss %.2f", epoch + 1, sum_epoch_loss / num_minibatches)
        train_metrics = self.neural_module.report_metrics()

        for epoch_hook in self.__train_epoch_end_hooks:
            epoch_hook(self.__model, self.neural_module, epoch, train_metrics)

        if len(train_metrics) > 0:
            self.LOGGER.info("Training Metrics: %s", json.dumps(train_metrics, indent=2))

    def _run_validation(
        self, validation_tensors, epoch, best_target_metric, device, parallelize, show_progress_bar
    ):
        self.neural_module.eval()
        sum_epoch_loss, num_minibatches, num_samples = 0.0, 0, 0
        start_time = time.time()
        with tqdm(
            desc="Validation", disable=not show_progress_bar, leave=False
        ) as progress_bar, torch.no_grad():
            for mb_data, raw_samples in self.__model.minibatch_iterator(
                validation_tensors(),
                device=device,
                max_minibatch_size=self.__minibatch_size,
                yield_partial_minibatches=True,
                shuffle_input=False,
                parallelize=parallelize,
            ):
                with torch.cuda.amp.autocast(enabled=self.__enable_amp):
                    mb_loss = self.neural_module(**mb_data)
                num_minibatches += 1
                num_samples += len(raw_samples)
                sum_epoch_loss += float(mb_loss.cpu())
                progress_bar.update()
                progress_bar.set_postfix(Loss=f"{sum_epoch_loss / num_minibatches:.2f}")

        elapsed_time = time.time() - start_time
        assert num_samples > 0, "No validation data was found."

        validation_loss = sum_epoch_loss / num_minibatches
        self.LOGGER.info(
            "Validation complete in %.1fsec [%.2f samples/sec]",
            elapsed_time,
            (num_samples / elapsed_time),
        )
        self.LOGGER.info("Epoch %i: Valid Loss %.2f", epoch + 1, validation_loss)

        validation_metrics = self.neural_module.report_metrics()
        for epoch_hook in self.__validation_epoch_end_hooks:
            epoch_hook(self.__model, self.neural_module, epoch, validation_metrics)
        if len(validation_metrics) > 0:
            self.LOGGER.info("Validation Metrics: %s", json.dumps(validation_metrics, indent=2))

        if self.__target_metric is not None:
            target_metric = validation_metrics[self.__target_metric]
        else:
            target_metric = validation_loss

        if self.__target_metric_higher_is_better:
            target_metric_improved = target_metric > best_target_metric
        else:
            target_metric_improved = target_metric < best_target_metric

        return target_metric, target_metric_improved

    def train(
        self,
        training_data: Iterable[TRawDatapoint],
        validation_data: Iterable[TRawDatapoint],
        *,
        show_progress_bar: bool = True,
        validate_on_start: bool = True,
        patience: int = 5,
        initialize_metadata: bool = True,
        parallelize: bool = True,
        use_multiprocessing: bool = True,
        exponential_running_average_factor: float = 0.97,
        device=None,
        store_tensorized_data_in_memory: bool = False,
        shuffle_training_data: bool = True,
    ) -> None:
        """
        The training-validation loop for `AbstractNeuralModel`s.

        :param training_data: An iterable that each iteration yields the full training data.
        :param validation_data: An iterable that each iteration yields the full validation data.
        :param show_progress_bar: Show a progress bar
        :param validate_on_start: Whether to run a validation loop on start
        :param patience: The number of iterations before early stopping kicks in.
        :param initialize_metadata: If true, initialize the metadata from the training_data. Otherwise,
            assume that the model that is being trained has its metadata already initialized.
        :param parallelize: Bool indicating whether to run in parallel
        :param use_multiprocessing: Whether to use multiprocessing
        :param exponential_running_average_factor: The factor of the running average of the training loss
            displayed in the progress bar.
        :param device: the target PyTorch device for training
        :param store_tensorized_data_in_memory: store all tensorized data in memory instead of computing them on-line.
        :param shuffle_training_data: shuffle the incoming data from `training_data`.
        """
        if initialize_metadata:
            self.load_metadata_and_create_network(training_data, parallelize, show_progress_bar)

        self.LOGGER.info(
            "Model has %s trainable parameters.",
            sum(
                param.numel()
                for param in self.neural_module.parameters(recurse=True)
                if param.requires_grad
            ),
        )

        training_tensors = lambda: self.__model.tensorize_dataset(
            iter(training_data), parallelize=parallelize, use_multiprocessing=use_multiprocessing
        )
        validation_tensors = lambda: self.__model.tensorize_dataset(
            iter(validation_data), parallelize=parallelize, use_multiprocessing=use_multiprocessing
        )
        if store_tensorized_data_in_memory:
            training_tensors = MemorizedDataIterable(training_tensors, shuffle=True)
            validation_tensors = MemorizedDataIterable(validation_tensors)

        if device is None:
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.LOGGER.info("Using `%s` for training." % device)
        self.neural_module.to(device)

        optimizer = self.__create_optimizer(self.neural_module.parameters())
        scheduler = None if self.__create_scheduler is None else self.__create_scheduler(optimizer)

        for hook in self.__training_start_hooks:
            hook(self.__model, self.neural_module, optimizer)

        if self.__target_metric_higher_is_better and self.__target_metric is not None:
            best_target_metric = -math.inf
        else:
            best_target_metric = math.inf

        if validate_on_start:
            target_metric, improved = self._run_validation(
                validation_tensors, 0, best_target_metric, device, parallelize, show_progress_bar
            )
            assert improved
            self.LOGGER.info(f"Initial {self.__target_metric or 'Loss'}: {target_metric}")
            best_target_metric = target_metric

        num_epochs_not_improved: int = 0
        for epoch in range(self.__max_num_epochs):
            self._run_training(
                training_tensors,
                epoch,
                device,
                exponential_running_average_factor,
                optimizer,
                parallelize,
                scheduler,
                show_progress_bar,
                shuffle_training_data,
            )

            target_metric, target_metric_improved = self._run_validation(
                validation_tensors,
                epoch,
                best_target_metric,
                device,
                parallelize,
                show_progress_bar,
            )
            if target_metric_improved:
                self.LOGGER.info(
                    f"Best performance so far "
                    f"({self.__target_metric or 'Loss'}: {target_metric:.3f} from {best_target_metric:.3f}). "
                    "Saving model checkpoint."
                )
                num_epochs_not_improved = 0
                self.__save_checkpoint()
                best_target_metric = target_metric
            else:
                num_epochs_not_improved += 1
                if num_epochs_not_improved > patience:
                    self.LOGGER.warning(
                        f"The target metric has not improved for {num_epochs_not_improved} epochs . Stopping."
                    )
                    break

        # Restore the best model params that were found.
        self.__restore_checkpoint(device=device)



import torch
from torch import nn
from typing import Any, Dict, Tuple

from ptgnn.baseneuralmodel import AbstractNeuralModel
from ptgnn.neuralmodels.embeddings.strelementrepresentationmodel import (
    StrElementRepresentationModel,
)
from ptgnn.neuralmodels.gnn.structs import AbstractNodeEmbedder


class CandidateNodeAnnotationModule(nn.Module):
    def __init__(self, node_embeddings_module):
        super().__init__()
        self.__node_embedding_module = node_embeddings_module

    def forward(self, node_data, is_candidate):
        embeddings = self.__node_embedding_module(**node_data)
        return torch.cat((embeddings, is_candidate.unsqueeze(-1)), dim=-1)


class CandidateNodeAnnotationModel(
    AbstractNeuralModel[Tuple[str, bool], Any, CandidateNodeAnnotationModule],
    AbstractNodeEmbedder,
):
    def __init__(self, embedding_size: int = 128, **kwargs):
        super().__init__()
        self.__str_node_annotation = StrElementRepresentationModel(
            embedding_size=embedding_size - 1, **kwargs
        )

    def update_metadata_from(self, datapoint: Tuple[str, bool]) -> None:
        self.__str_node_annotation.update_metadata_from(datapoint[0])

    def build_neural_module(self) -> CandidateNodeAnnotationModule:
        return CandidateNodeAnnotationModule(
            node_embeddings_module=self.__str_node_annotation.build_neural_module(),
        )

    def tensorize(self, datapoint: Tuple[str, bool]) -> Tuple[Any, bool]:
        return self.__str_node_annotation.tensorize(datapoint[0]), datapoint[1]

    def initialize_minibatch(self) -> Dict[str, Any]:
        return {"node_data": self.__str_node_annotation.initialize_minibatch(), "is_candidate": []}

    def extend_minibatch_with(
        self, tensorized_datapoint: Tuple[Any, bool], partial_minibatch: Dict[str, Any]
    ) -> bool:
        continue_extending = self.__str_node_annotation.extend_minibatch_with(
            tensorized_datapoint[0], partial_minibatch["node_data"]
        )
        partial_minibatch["is_candidate"].append(tensorized_datapoint[1])
        return continue_extending

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Any
    ) -> Dict[str, Any]:
        return {
            "node_data": self.__str_node_annotation.finalize_minibatch(
                accumulated_minibatch_data["node_data"], device=device
            ),
            "is_candidate": torch.tensor(
                accumulated_minibatch_data["is_candidate"], dtype=torch.float32, device=device
            ),
        }

    def representation_size(self) -> int:
        return self.__str_node_annotation.representation_size() + 1



import numpy as np
import torch
from abc import ABC, abstractmethod
from typing import Dict, Generic, List, NamedTuple, Optional, Tuple, TypeVar

TNodeData = TypeVar("TNodeData")
TEdgeData = TypeVar("TEdgeData")
TTensorizedNodeData = TypeVar("TTensorizedNodeData")
TTensorizedEdgeData = TypeVar("TTensorizedEdgeData")


class GraphData(Generic[TNodeData, TEdgeData]):
    __slots__ = ("node_information", "edges", "edge_features", "reference_nodes")

    def __init__(
        self,
        node_information: List[TNodeData],
        edges: Dict[str, List[Tuple[int, int]]],
        reference_nodes: Dict[str, List[int]],
        edge_features: Optional[Dict[str, List[TEdgeData]]] = None,
    ):
        self.node_information = node_information
        self.edges = edges
        self.edge_features = edge_features
        self.reference_nodes = reference_nodes


class TensorizedGraphData(Generic[TTensorizedNodeData, TTensorizedEdgeData]):
    __slots__ = (
        "num_nodes",
        "node_tensorized_data",
        "adjacency_lists",
        "edge_features",
        "reference_nodes",
    )

    def __init__(
        self,
        num_nodes: int,
        node_tensorized_data: List[TTensorizedNodeData],
        adjacency_lists: List[Tuple[np.ndarray, np.ndarray]],
        edge_features: Optional[List[TTensorizedEdgeData]],
        reference_nodes: Dict[str, np.ndarray],
    ):
        self.num_nodes = num_nodes
        self.node_tensorized_data = node_tensorized_data
        self.adjacency_lists = adjacency_lists
        self.edge_features = edge_features
        self.reference_nodes = reference_nodes


class GnnOutput(NamedTuple):
    input_node_representations: torch.Tensor
    output_node_representations: torch.Tensor
    node_to_graph_idx: torch.Tensor
    # Which are the idxs of the referenced nodes (in the minibatch)?
    node_idx_references: Dict[str, torch.Tensor]
    # Which graph do the referenced nodes belong in?
    node_graph_idx_reference: Dict[str, torch.Tensor]
    num_graphs: int


class AbstractNodeEmbedder(ABC):
    """Abstract node embedder."""

    @abstractmethod
    def representation_size(self) -> int:
        pass



from typing_extensions import final

import gzip
import os
import torch
from abc import ABC, abstractmethod
from concurrent import futures
from dpu_utils.utils.iterators import BufferedIterator, ThreadedIterator, shuffled_iterator
from itertools import islice
from pathlib import Path
from torch import nn
from typing import (
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

__all__ = ["AbstractNeuralModel"]

TRawDatapoint = TypeVar("TRawDatapoint")
TTensorizedDatapoint = TypeVar("TTensorizedDatapoint")
TNeuralModule = TypeVar("TNeuralModule", bound=nn.Module)
T = TypeVar("T")


class AbstractNeuralModel(ABC, Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
    """
    This class provides the structure for defining neural networks in a compositional way.
    An `AbstractNeuralModel` accepts as input `TRawDatapoint`, a single example. Each model can have some
    metadata. For example, this may be the vocabulary of words to be represented and their mappings to integer ids.
    The computed metadata is then used to build a `TNeuralModule` (a `nn.Module`) which contains
    the definition of the neural network architecture and its parameters. During training and testing, each `TRawDatapoint`
    is converted to a tensorized format `TTensorizedDatapoint`. By batching multiple `TTensorizedDatapoint` a minibatch
    is created which is the input of the `TNeuralModule`.

    To create an `AbstractNeuralModel`:

    * Compute the metadata.
        * Initialize metadata in `initialize_metadata`. Store any metadata structures within the model.
        * Update this structure for each example in `update_metadata_from`. Explicitly unpack/transform any parts of the
            `TRawDatapoint` and invoke the `update_metadata_from` for all child models.
        * Finally, compute the model metadata in `finalize_metadata` and store them as fields in your implementation
            of the `AbstractNeuralModel`. Optionally, discard any temporary structures within the model.

    * Implement `build_neural_module()` that uses the metadata previously computed to create a `TNeuralModule`, ie.
        a concrete `nn.Module` or `ModuleWithMetrics`. Note that no elements of the neural network should be stored
        in this object but they should be returned.
    * Define `tensorize()` accepting a single `TRawDatapoint` and yielding a `TTensorizedDatapoint`.
        Each `TTensorizedDatapoint` should be considered a single sample for your neural network. Each model should
        call its child models (if any) to delegate tensorization of the relevant parts of the `TRawDatapoint`

    * Finally, to create minibatches:
        * Implement `initialize_minibatch` which creates a dictionary that will be populated with the information from
            multiple `TTensorizedDatapoint`s. Explicitly invoke child models to initialize aspect of the minibatch
            that are their responsibility.
        * Implement `extend_minibatch_with` to extend the minibatch with an additional sample. Return `True`
            if the minibatch can be further extended. Explicitly deconstruct the `TTensorizedDatapoint` and pass its
            parts to the responsible child models.
        * Implement `finalize_minibatch` to finalize the minibatch (invoking this for any child models). This
            should return a dictionary of PyTorch tensors ready to be consumed by the `forward()` of the `TNeuralModule`.
    """

    def __init__(self):
        self.__metadata_initialized = False

    @final
    @property
    def model_definition(self) -> Mapping[str, Any]:
        """A description of this model."""
        description: Dict[str, Any] = {}
        for attr, value in self.__dict__.items():
            if isinstance(value, AbstractNeuralModel):
                description[attr] = value.model_definition
            elif isinstance(value, (int, float, str, bool, NamedTuple)):
                description[attr] = value
        return description

    # region Metadata and Neural Network Building
    def initialize_metadata(self) -> None:
        """
        Optionally initialize any model metadata (children metadata will be initialized separately).
        """

    @abstractmethod
    def update_metadata_from(self, datapoint: TRawDatapoint) -> None:
        """
        Update the model metadata from a single `datapoint`.

        Implementors of this method should:
        * Update all metadata that is needed by the model.
        * Unpack any parts of the parts of `datapoint` that are needed for the child models
            and invoke here their `update_metadata_from`.
        """
        raise NotImplementedError()

    def finalize_metadata(self) -> None:
        """
        Compute the final metadata. The final metadata should be stored within the model.
        Optionally, discard any temporary structures used to compute the finalized metadata.
        """

    def __initialize_metadata_recursive(self) -> None:
        self.initialize_metadata()
        for attr, value in self.__dict__.items():
            if isinstance(value, AbstractNeuralModel):
                value.__initialize_metadata_recursive()

    def __finalize_metadata_recursive(self) -> None:
        self.finalize_metadata()
        for attr, value in self.__dict__.items():
            if isinstance(value, AbstractNeuralModel):
                value.__finalize_metadata_recursive()
        self.__metadata_initialized = True

    @final
    def compute_metadata(
        self, dataset_iterator: Iterator[TRawDatapoint], parallelize: bool = True
    ) -> None:
        """
        Compute the metadata for this model including its children.
        This function should be invoked by the root-level model.
        """
        assert not self.__metadata_initialized, "Metadata has already been initialized."
        self.__initialize_metadata_recursive()
        for element in ThreadedIterator(dataset_iterator, enabled=parallelize):
            self.update_metadata_from(element)
        self.__finalize_metadata_recursive()

    @abstractmethod
    def build_neural_module(self) -> TNeuralModule:
        """
        Create the neural network that corresponds to this model.

        Note to implementors:
            * All model metadata will have been computed by this point and can be used here.
            * The output neural network should *not* be stored in this object but should be returned.
            * Invoke build_neural_module() for all child models and then compose a single neural module
              (ie. a `nn.Module` or `ModuleWithMetrics`) that encapsulates the children modules.
        """
        raise NotImplementedError()

    # endregion

    # region Saving/Loading
    def save(self, path: Path, model: TNeuralModule) -> None:
        os.makedirs(os.path.dirname(str(path.absolute())), exist_ok=True)
        with gzip.open(path, "wb") as f:
            torch.save((self, model), f)

    @classmethod
    def restore_model(cls: Type[T], path: Path, device=None) -> Tuple[T, TNeuralModule]:
        with gzip.open(path, "rb") as f:
            return torch.load(f, map_location=device)

    # endregion

    # region Tensor Conversion
    @abstractmethod
    def tensorize(self, datapoint: TRawDatapoint) -> Optional[TTensorizedDatapoint]:
        """
        This is called to tensorize the data of a single input example in a form that can be consumed by the
        neural network. This function yields a `TTensorizedDatapoint` or `None` if this datapoint should be
        discarded.

        Note to implementors: this usually involves unpacking the datapoint and invoking child models'
          `tensorize()` with the portions of the data that each child model is responsible about. Finally,
          this function should encapsulate this data into a `TTensorizedDatapoint` along with any additional
          tensorized information needed by this model.
        """
        raise NotImplementedError()

    @final
    def tensorize_dataset(
        self,
        dataset_iterator: Iterator[TRawDatapoint],
        *,
        parallelize: bool = True,
        use_multiprocessing: bool = True,
        return_input_data: bool = False,
    ) -> Iterator[Tuple[TTensorizedDatapoint, Optional[TRawDatapoint]]]:
        """
        Given a dataset, return an iterator of the tensorized data. If `return_input_data` is True
        then the original data is also returned.

        :param dataset_iterator: An iterator of raw data.
        :param parallelize: If True, the tensorization will be parallelized. This speeds up tensorization but
            makes debugging harder.
        :param use_multiprocessing: If parallelize==True, then use multiprocessing instead of multithreading.
            If parallelize==False this option is ignored.
        :param return_input_data: Should the raw input data also be returned?
        """
        assert self.__metadata_initialized, "Metadata has not been initialized."
        if parallelize:
            base_iterator = (
                (self.tensorize(d), d if return_input_data else None) for d in dataset_iterator
            )
            if use_multiprocessing:
                for tensorized_sample in BufferedIterator(base_iterator):
                    if tensorized_sample[0] is not None:
                        yield tensorized_sample
            else:
                with futures.ThreadPoolExecutor() as pool:
                    for tensorized_sample in pool.map(
                        lambda d: (
                            self.tensorize(d),
                            d if return_input_data else None,
                        ),
                        dataset_iterator,
                        chunksize=20,
                    ):
                        if tensorized_sample[0] is not None:
                            yield tensorized_sample
        else:
            for datapoint in dataset_iterator:
                tensorized_data_sample = self.tensorize(datapoint)
                if tensorized_data_sample is not None:
                    yield (tensorized_data_sample, datapoint if return_input_data else None)

    # endregion

    # region Minibatching Logic
    @abstractmethod
    def initialize_minibatch(self) -> Dict[str, Any]:
        """
        Initialize a dictionary that will be incrementally populated by `extend_minibatch_with()`. Once the minibatch
        is full, `finalize_minibatch()` will be invoked.

        Explicitly invoke `initialize_minibatch` for child models to initialize the parts of minibatch that are
         their responsibility.
        """
        raise NotImplementedError()

    @abstractmethod
    def extend_minibatch_with(
        self, tensorized_datapoint: TTensorizedDatapoint, partial_minibatch: Dict[str, Any]
    ) -> bool:
        """
        Add a datapoint to the partial minibatch. If for some reason the minibatch cannot accumulate
            additional samples after this one, this method should return False. Once the minibatch
            is full, `finalize_minibatch()` will be invoked.

        Note to implementors: Commonly, each model will unpack the `tensorized_datapoint` and invoke `extend_minibatch_with`
            of its child models. Then, if at least one of the child models returns False
            then this function should also return False.

        :param tensorized_datapoint: the datapoint to be added. This parameter is the output of `tensorize()`.
        :param partial_minibatch: the minibatch data to be populated. This is the dictionary that was initialized
            by `initialize_minibatch`.
        :return True if the minibatch can be further extended. False if for some reason the minibatch is full.
        """
        raise NotImplementedError()

    @abstractmethod
    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        """
        Finalize the minibatch data and make sure that the data is in an appropriate format to be consumed by
        the model. Commonly the values of the returned dictionary are `torch.tensor()` and the keys of the
        returned dictionary are the parameters of the neural module's forward().

        :param accumulated_minibatch_data: the data accumulated by `extend_minibatch_with`.
        :param device: the device for the torch tensors.
        :return: a dictionary `mb_data` that will is the input to the `forward(**mb_data)` of the model's neural module.
        """
        raise NotImplementedError()

    @final
    def __iterate_unfinalized_minibatches(
        self,
        tensorized_data: Iterator[Tuple[TTensorizedDatapoint, Optional[TRawDatapoint]]],
        max_minibatch_size: int,
        yield_partial_minibatches: bool = True,
    ) -> Iterator[Tuple[Dict[str, Any], List[Optional[TRawDatapoint]]]]:
        """
        Initialize and accumulate data into minibatches, but does not finalize them (so that this process can
         be offloaded to a different thread/process).
        """
        while True:
            mb_data = self.initialize_minibatch()
            mb_input_data, num_elements_added = [], 0
            for tensorized_sample, input_data in islice(tensorized_data, max_minibatch_size):
                continue_extending = self.extend_minibatch_with(tensorized_sample, mb_data)
                mb_input_data.append(input_data)
                num_elements_added += 1
                if not continue_extending:
                    # The implementation of the model asked to stop extending the minibatch.
                    data_iterator_exhausted = False
                    break
            else:
                # The data is exhausted if we finished iterating through the loop and still don't have max_num_items
                data_iterator_exhausted = num_elements_added < max_minibatch_size

            if num_elements_added == 0 or (
                data_iterator_exhausted and not yield_partial_minibatches
            ):
                return
            yield mb_data, mb_input_data

    @final
    def minibatch_iterator(
        self,
        tensorized_data: Iterator[Tuple[TTensorizedDatapoint, Optional[TRawDatapoint]]],
        device: Union[str, torch.device],
        max_minibatch_size: int,
        yield_partial_minibatches: bool = True,
        shuffle_input: bool = False,
        parallelize: bool = True,
    ) -> Iterator[Tuple[Dict[str, Any], List[Optional[TRawDatapoint]]]]:
        """
        An iterator that yields minibatches to be consumed by a neural module.
        :param tensorized_data: An iterator of tensorized data. Commonly that's the output of tensorize_dataset()
        :param device: The device on which the tensorized data will be stored.
        :param max_minibatch_size: the maximum size of the minibatch.
        :param yield_partial_minibatches: If true, yield partial minibatches, i.e. minibatches that do not
            reach the `max_minibatch_size` and the `extend_minibatch_with` did not consider full.
            Users might want to set this to False, when training.
        :param shuffle_input: Should the `tensorized_data` be shuffled? (e.g. during training)
        :param parallelize: if True, minibatching will be parallelized. This may make debugging harder.
        :return: an iterator that yield tuples, with the minibatch data and the raw data points (if they are present in `tensorized_data`).
        """
        assert self.__metadata_initialized, "Metadata has not been initialized."

        if shuffle_input:
            tensorized_data = shuffled_iterator(tensorized_data)

        unfinalized_minibatches = ThreadedIterator(
            self.__iterate_unfinalized_minibatches(
                tensorized_data, max_minibatch_size, yield_partial_minibatches
            ),
            enabled=parallelize,
        )
        yield from ThreadedIterator(
            ((self.finalize_minibatch(d[0], device), d[1]) for d in unfinalized_minibatches),
            enabled=parallelize,
        )

    # endregion



#!/usr/bin/env python
"""
Usage:
    predict.py [options] MODEL_FILENAME DATA_PATH

Options:
    --azure-info=<path>        Azure authentication information file (JSON). Used to load data from Azure storage.
    -h --help                  Show this screen.
    --debug                    Enable debug routines. [default: False]
"""
from docopt import docopt
from dpu_utils.utils import RichPath, run_and_debug
from pathlib import Path

from ptgnn.implementations.typilus.graph2class import Graph2Class
from ptgnn.implementations.typilus.train import load_from_folder


def run(arguments):
    azure_info_path = arguments.get("--azure-info", None)
    data_path = RichPath.create(arguments["DATA_PATH"], azure_info_path)
    data = load_from_folder(data_path, shuffle=False)

    model_path = Path(arguments["MODEL_FILENAME"])
    model, nn = Graph2Class.restore_model(model_path, "cpu")

    predictions = model.predict(data, nn, "cpu")
    for graph, suggestions in predictions:
        for supernode_idx, (target_type, prob) in suggestions.items():
            supernode_info = graph["supernodes"][str(supernode_idx)]
            print(
                f'`{supernode_info["name"]}` Original: `{supernode_info["annotation"]}` Predicted: `{target_type}` ({prob:.2%})'
            )


if __name__ == "__main__":
    args = docopt(__doc__)
    run_and_debug(lambda: run(args), args.get("--debug", False))



from typing_extensions import final

from abc import ABC
from torch import nn
from typing import Any, Dict


class ModuleWithMetrics(nn.Module, ABC):
    """
    A decorator around `nn.Module` that allows a module and its ancestors to collect and report metrics, such
    as accuracy, loss, etc.

    Each Module can collect arbitrary metrics in its fields. The Module implementors should initialize
    and update the metrics as they see fit. Implementors should define
     * `_reset_module_metrics` that resets the module metrics (e.g. sets a counter to zero)
     * `_module_metrics` that reports a dictionary of metrics for this specific module.

    `ModuleWithMetrics` can recursively collect all the metrics from its children and report it
    by invoking `report_metrics()`. Metrics are automatically reset when `train()` or `eval()`
    is invoked.
    """

    def __init__(self):
        super().__init__()
        self._reset_module_metrics()

    @final
    def report_metrics(self) -> Dict[str, Any]:
        """
        Report the collected metrics for this `ModuleWithMetrics` and its descendant modules.

        Each module can internally collect its own metrics as the implementor sees fit. For example,
        a counter may be incremented when the `forward()` function is invoked or a running average may
        by updated when a loss is computed. The metrics counter can be reset outside of the module
        when `reset_metrics` is invoked.

        To add metrics to a Module, implementors need to:
        * Implement `_module_metrics` that computes the reported metrics from any component-internal variables.
        * Implement `_reset_module_metrics` which resets any variables that compute metrics.
        * Store any metric-related variables as fields in the module.
        """
        metrics = self._module_metrics()
        for child_module in self.modules():
            if isinstance(child_module, ModuleWithMetrics):
                child_metrics = child_module._module_metrics()
                if len(child_metrics) > 0:
                    metrics.update(child_metrics)
        return metrics

    @final
    def reset_metrics(self) -> None:
        """Reset any reported metrics. Often called after report_metrics() to reset any counters etc."""
        self._reset_module_metrics()
        for child_module in self.modules():
            if isinstance(child_module, ModuleWithMetrics):
                child_module._reset_module_metrics()

    def train(self, mode: bool = True) -> "ModuleWithMetrics":
        self.reset_metrics()
        return super().train(mode=mode)

    def eval(self) -> "ModuleWithMetrics":
        self.reset_metrics()
        return super().eval()

    def _module_metrics(self) -> Dict[str, Any]:
        """
        Return a dictionary of metrics for the current module.

        The key is the name of the metric as it will be appear reported.
        The value can be anything, but using a formatted string may often be the preferred choice.
        """
        return {}

    def _reset_module_metrics(self) -> None:
        """Reset any metrics related to the module, such as any counters, running sums, averages, etc."""
        pass



from typing_extensions import Final

import logging
import numpy as np
import torch
from collections import defaultdict
from torch import nn
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from ptgnn.baseneuralmodel import AbstractNeuralModel, ModuleWithMetrics
from ptgnn.baseneuralmodel.utils.data import enforce_not_None
from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import AbstractMessagePassingLayer
from ptgnn.neuralmodels.gnn.structs import GnnOutput, GraphData, TensorizedGraphData


class GraphNeuralNetwork(ModuleWithMetrics):
    """
    A generic message-passing graph neural network with discrete edge types.
    """

    def __init__(
        self,
        message_passing_layers: List[AbstractMessagePassingLayer],
        node_embedder: nn.Module,
        introduce_backwards_edges: bool,
        add_self_edges: bool,
        edge_dropout_rate: float = 0.0,
        edge_feature_embedder: Optional[nn.Module] = None,
    ):
        """
        :param message_passing_layers: A list of message passing layers.
        :param node_embedder: a `nn.Module` that converts node data into a vector representation
        :param introduce_backwards_edges: If `True` special backwards edges should be automatically created.
        :param add_self_edges: If `True` self-edges will be added. These edges connect the same node across
            multiple timesteps.
        :param edge_dropout_rate: remove random pct of edges
        """
        super().__init__()
        self.__message_passing_layers = nn.ModuleList(message_passing_layers)
        self.__node_embedder = node_embedder
        self.__introduce_backwards_edges = introduce_backwards_edges
        self.__add_self_edges = add_self_edges
        assert 0 <= edge_dropout_rate < 1
        self.__edge_dropout_rate = edge_dropout_rate
        self.__edge_feature_embedder = edge_feature_embedder

    @property
    def input_node_state_dim(self) -> int:
        """The dimension of the input GNN node states."""
        return self.__message_passing_layers[0].input_state_dimension

    @property
    def output_node_state_dim(self) -> int:
        """The dimension of the output GNN node states."""
        return self.__message_passing_layers[-1].output_state_dimension

    @property
    def message_passing_layers(self) -> List[AbstractMessagePassingLayer]:
        return self.__message_passing_layers

    def _reset_module_metrics(self) -> None:
        self.__num_graphs, self.__num_edges, self.__num_nodes = 0, 0, 0

    def _module_metrics(self) -> Dict[str, Any]:
        return {
            "num_graphs": self.__num_graphs,
            "num_nodes": self.__num_nodes,
            "num_edges": self.__num_edges,
        }

    def gnn(
        self,
        node_representations: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        edge_feature_embeddings: List[torch.Tensor],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        return_all_states: bool = False,
    ) -> torch.Tensor:
        """
        :param node_representations: A [num_nodes, hidden_dimension] matrix of node representations.
        :param adjacency_lists: a list of [num_edges_per_type, 2] adjacency lists per edge type.
                The order is fixed across runs. Backwards edges and self-edges are included if
                the appropriate hyperparameter is set.
        :param edge_feature_embeddings: a list of the edge features per edge-type.
        :param node_to_graph_idx: A mapping that tells us which graph the node belongs to
        :param reference_node_ids: A dictionary indicating the reference node index
        :param reference_node_graph_idx: A dictionary indicating the graph index for reference node
        :param return_all_states: Whether to return all states
        :return: a [num_nodes, output_hidden_dimension] matrix of the output representations
        """
        if self.__edge_dropout_rate > 0 and self.training:
            dropped_adj_list, dropped_edge_features = [], []
            for (edge_sources_idxs, edge_target_idxs), edge_features in zip(
                adjacency_lists, edge_feature_embeddings
            ):
                mask = (
                    torch.rand_like(edge_sources_idxs, dtype=torch.float32)
                    > self.__edge_dropout_rate
                )
                dropped_adj_list.append(
                    (edge_sources_idxs.masked_select(mask), edge_target_idxs.masked_select(mask))
                )
                dropped_edge_features.append(edge_features[mask])
            adjacency_lists = dropped_adj_list
            edge_feature_embeddings = dropped_edge_features

        all_states = [node_representations]
        for mp_layer_idx, mp_layer in enumerate(self.__message_passing_layers):
            node_representations = mp_layer(
                node_states=node_representations,
                adjacency_lists=adjacency_lists,
                node_to_graph_idx=node_to_graph_idx,
                reference_node_ids=reference_node_ids,
                reference_node_graph_idx=reference_node_graph_idx,
                edge_features=edge_feature_embeddings,
            )
            all_states.append(node_representations)
        if return_all_states:
            node_representations = torch.cat(all_states, dim=-1)
        return node_representations

    def forward(
        self,
        *,
        node_data,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        edge_feature_data: List,
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        num_graphs,
        **kwargs
    ) -> GnnOutput:
        """

        :param node_data: The data for the node embedder to compute the initial node representations.
        :param adjacency_lists: A list of [num_edges, 2] matrices for each edge type.
        :param edge_feature_data: A list of the same size as `adjacency_lists` with the data
            to compute the edge features.
        :param node_to_graph_idx: A [num_nodes] vector that contains the index of the graph it belongs to.
        :param reference_node_ids: A dictionary with values the indices of the reference nodes.
        :param reference_node_graph_idx: A dictionary with values the index of the graph that each
            node in the `reference_node_ids` belongs to.
        :param num_graphs: the number of graphs.
        """
        initial_node_representations = self.__node_embedder(**node_data)  # [num_nodes, D]

        if self.__edge_feature_embedder is None:
            edge_feature_embeddings = [
                torch.empty(f.shape[0], 0, device=node_to_graph_idx.device)
                for f, _ in adjacency_lists
            ]
        else:
            edge_feature_embeddings = [
                self.__edge_feature_embedder(**edge_data) for edge_data in edge_feature_data
            ]

        if self.__introduce_backwards_edges:
            adjacency_lists += [(t, f) for f, t in adjacency_lists]
            edge_feature_embeddings += [e for e in edge_feature_embeddings]

        if self.__add_self_edges:
            num_nodes = node_to_graph_idx.shape[0]
            idents = torch.arange(num_nodes, dtype=torch.int64, device=node_to_graph_idx.device)
            adjacency_lists.append((idents, idents))
            edge_feature_embeddings.append(
                torch.zeros(
                    num_nodes,
                    edge_feature_embeddings[-1].shape[-1],
                    device=node_to_graph_idx.device,
                )
            )

        output_representations = self.gnn(
            initial_node_representations,
            adjacency_lists,
            edge_feature_embeddings,
            node_to_graph_idx,
            reference_node_ids,
            reference_node_graph_idx,
            **kwargs
        )  # [num_nodes, H]

        with torch.no_grad():
            self.__num_edges += int(sum(adj[0].shape[0] for adj in adjacency_lists))
            self.__num_graphs += int(num_graphs)
            self.__num_nodes += int(node_to_graph_idx.shape[0])
        return GnnOutput(
            input_node_representations=initial_node_representations,
            output_node_representations=output_representations,
            node_to_graph_idx=node_to_graph_idx,
            node_idx_references=reference_node_ids,
            node_graph_idx_reference=reference_node_graph_idx,
            num_graphs=num_graphs,
        )


TNodeData = TypeVar("TNodeData")
TEdgeData = TypeVar("TEdgeData")
TTensorizedNodeData = TypeVar("TTensorizedNodeData")
TTensorizedEdgeData = TypeVar("TTensorizedEdgeData")


class GraphNeuralNetworkModel(
    AbstractNeuralModel[
        GraphData[TNodeData, TEdgeData],
        TensorizedGraphData[TTensorizedNodeData, TTensorizedEdgeData],
        GraphNeuralNetwork,
    ],
):
    LOGGER: Final = logging.getLogger(__name__)

    def __init__(
        self,
        *,
        node_representation_model: AbstractNeuralModel[TNodeData, TTensorizedNodeData, nn.Module],
        message_passing_layer_creator: Callable[[int], List[AbstractMessagePassingLayer]],
        max_nodes_per_graph: int = 80000,
        max_graph_edges: int = 100000,
        introduce_backwards_edges: bool = True,
        stop_extending_minibatch_after_num_nodes: int = 10000,
        add_self_edges: bool = False,
        edge_dropout_rate: float = 0.0,
        edge_representation_model: Optional[
            AbstractNeuralModel[TEdgeData, TTensorizedEdgeData, nn.Module]
        ] = None
    ):
        """
        :param node_representation_model: A model that can convert the data of each node into their
            tensor representation and create the neural network that computes the node representations.
        :param message_passing_layer_creator: A function that accepts the number of edge types and creates
            a list of `AbstractMessagePassingLayer` layers to be used by the GNN.
        """
        super().__init__()
        self.__message_passing_layers_creator: Final = message_passing_layer_creator
        self.__node_embedding_model: Final = node_representation_model
        self.__edge_embedding_model: Final = edge_representation_model
        self.max_nodes_per_graph: Final = max_nodes_per_graph
        self.max_graph_edges: Final = max_graph_edges
        self.introduce_backwards_edges: Final = introduce_backwards_edges
        self.stop_extending_minibatch_after_num_nodes: Final = (
            stop_extending_minibatch_after_num_nodes
        )
        self.add_self_edges: Final = add_self_edges
        self.__edge_dropout_rate = edge_dropout_rate

    # region Metadata Loading
    def initialize_metadata(self) -> None:
        self.__edge_types_mdata: Set[str] = set()

    def update_metadata_from(self, datapoint: GraphData[TNodeData, TEdgeData]) -> None:
        for node in datapoint.node_information:
            self.__node_embedding_model.update_metadata_from(node)

        for edge_type in datapoint.edges:
            self.__edge_types_mdata.add(edge_type)

        if datapoint.edge_features is not None and self.__edge_embedding_model is not None:
            for edge_features in datapoint.edge_features.values():
                for edge_feature in edge_features:
                    self.__edge_embedding_model.update_metadata_from(edge_feature)

    def finalize_metadata(self) -> None:
        self.LOGGER.info("Found %s edge types in data.", len(self.__edge_types_mdata))
        self.__edge_idx_to_type = tuple(self.__edge_types_mdata)
        self.__edge_types = {e: i for i, e in enumerate(self.__edge_idx_to_type)}
        del self.__edge_types_mdata

    @property
    def _num_edge_types(self) -> int:
        num_types = len(self.__edge_types)
        if self.introduce_backwards_edges:
            num_types *= 2
        if self.add_self_edges:
            num_types += 1
        return num_types

    def build_neural_module(self) -> GraphNeuralNetwork:
        if self.__edge_embedding_model is None:
            edge_feature_embedder = None
        else:
            edge_feature_embedder = self.__edge_embedding_model.build_neural_module()

        gnn = GraphNeuralNetwork(
            self.__message_passing_layers_creator(self._num_edge_types),
            node_embedder=self.__node_embedding_model.build_neural_module(),
            introduce_backwards_edges=self.introduce_backwards_edges,
            add_self_edges=self.add_self_edges,
            edge_dropout_rate=self.__edge_dropout_rate,
            edge_feature_embedder=edge_feature_embedder,
        )
        del self.__message_passing_layers_creator
        return gnn

    # endregion

    def edge_idx_by_name(self, name: str) -> int:
        return self.__edge_types[name]

    def __iterate_edge_types(
        self, data_to_load: GraphData[TNodeData, TEdgeData]
    ) -> Iterator[Tuple[np.ndarray, np.ndarray]]:
        for edge_type in self.__edge_idx_to_type:
            adjacency_list = data_to_load.edges.get(edge_type)
            if adjacency_list is not None and len(adjacency_list) > 0:
                adj = np.array(adjacency_list, dtype=np.int32)
                yield adj[:, 0], adj[:, 1]
            else:
                yield np.zeros(0, dtype=np.int32), np.zeros(0, dtype=np.int32)

    def tensorize(
        self, datapoint: GraphData[TNodeData, TEdgeData]
    ) -> Optional[TensorizedGraphData[TTensorizedNodeData, TTensorizedEdgeData]]:
        if len(datapoint.node_information) > self.max_nodes_per_graph:
            self.LOGGER.warning("Dropping graph with %s nodes." % len(datapoint.node_information))
            return None

        if self.__edge_embedding_model is None:
            tensorized_edge_features = None
        else:
            tensorized_edge_features = []
            for edge_type in self.__edge_idx_to_type:
                edge_features_for_edge_type = datapoint.edge_features.get(edge_type)
                if edge_features_for_edge_type is None:
                    # No edges of type `edge_type`
                    tensorized_edge_features.append([])
                else:
                    tensorized_edge_features.append(
                        [
                            self.__edge_embedding_model.tensorize(e)
                            for e in edge_features_for_edge_type
                        ]
                    )

        tensorized_data = TensorizedGraphData(
            adjacency_lists=list(self.__iterate_edge_types(datapoint)),
            node_tensorized_data=[
                enforce_not_None(self.__node_embedding_model.tensorize(ni))
                for ni in datapoint.node_information
            ],
            edge_features=tensorized_edge_features,
            reference_nodes={
                n: np.array(np.array(refs, dtype=np.int32))
                for n, refs in datapoint.reference_nodes.items()
            },
            num_nodes=len(datapoint.node_information),
        )

        num_edges = sum(len(adj) for adj in tensorized_data.adjacency_lists)
        if num_edges > self.max_graph_edges:
            self.LOGGER.warning("Dropping graph with %s edges." % num_edges)
            return None

        return tensorized_data

    # region Minibatching
    def initialize_minibatch(self) -> Dict[str, Any]:
        return {
            "node_data_mb": self.__node_embedding_model.initialize_minibatch(),
            "adjacency_lists": [([], []) for _ in range(len(self.__edge_types))],
            "edge_feature_data": [
                self.__edge_embedding_model.initialize_minibatch()
                if self.__edge_embedding_model is not None
                else None
                for _ in range(len(self.__edge_types))
            ],
            "num_nodes_per_graph": [],
            "reference_node_graph_idx": defaultdict(list),
            "reference_node_ids": defaultdict(list),
            "num_nodes_in_mb": 0,
        }

    def extend_minibatch_with(
        self,
        tensorized_datapoint: TensorizedGraphData[TTensorizedNodeData, TTensorizedEdgeData],
        partial_minibatch: Dict[str, Any],
    ) -> bool:
        continue_extending = True
        for node_tensorized_info in tensorized_datapoint.node_tensorized_data:
            continue_extending &= self.__node_embedding_model.extend_minibatch_with(
                node_tensorized_info, partial_minibatch["node_data_mb"]
            )

        graph_idx = len(partial_minibatch["num_nodes_per_graph"])

        adj_list = partial_minibatch["adjacency_lists"]
        tensorized_edge_feature_data = partial_minibatch["edge_feature_data"]
        nodes_in_mb_so_far = partial_minibatch["num_nodes_in_mb"]

        datapoint_edge_features = tensorized_datapoint.edge_features
        if datapoint_edge_features is None:
            datapoint_edge_features = [None for _ in range(len(adj_list))]

        for (
            sample_adj_list_for_edge_type,
            edge_features,
            mb_adj_lists_for_edge_type,
            mb_edge_feature_data,
        ) in zip(
            tensorized_datapoint.adjacency_lists,
            datapoint_edge_features,
            adj_list,
            tensorized_edge_feature_data,
        ):
            mb_adj_lists_for_edge_type[0].append(
                sample_adj_list_for_edge_type[0] + nodes_in_mb_so_far
            )
            mb_adj_lists_for_edge_type[1].append(
                sample_adj_list_for_edge_type[1] + nodes_in_mb_so_far
            )
            if self.__edge_embedding_model is not None:
                for edge_feature in edge_features:
                    self.__edge_embedding_model.extend_minibatch_with(
                        edge_feature, mb_edge_feature_data
                    )

        for ref_name, ref_nodes in tensorized_datapoint.reference_nodes.items():
            partial_minibatch["reference_node_graph_idx"][ref_name].extend(
                graph_idx for _ in range(len(ref_nodes))
            )
            partial_minibatch["reference_node_ids"][ref_name].append(ref_nodes + nodes_in_mb_so_far)

        partial_minibatch["num_nodes_per_graph"].append(tensorized_datapoint.num_nodes)
        partial_minibatch["num_nodes_in_mb"] = nodes_in_mb_so_far + tensorized_datapoint.num_nodes
        return partial_minibatch["num_nodes_in_mb"] < self.stop_extending_minibatch_after_num_nodes

    @staticmethod
    def __create_node_to_graph_idx(num_nodes_per_graph: List[int]) -> Iterable[int]:
        for i, graph_size in enumerate(num_nodes_per_graph):
            yield from (i for _ in range(graph_size))

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:

        if self.__edge_embedding_model is None:
            edge_feature_data = [None for _ in accumulated_minibatch_data["edge_feature_data"]]
        else:
            edge_feature_data = [
                self.__edge_embedding_model.finalize_minibatch(edge_features_for_type, device)
                for edge_features_for_type in accumulated_minibatch_data["edge_feature_data"]
            ]

        return {
            "node_data": self.__node_embedding_model.finalize_minibatch(
                accumulated_minibatch_data["node_data_mb"], device
            ),
            "adjacency_lists": [
                (
                    torch.tensor(np.concatenate(adjFrom), dtype=torch.int64, device=device),
                    torch.tensor(np.concatenate(adjTo), dtype=torch.int64, device=device),
                )
                for adjFrom, adjTo in accumulated_minibatch_data["adjacency_lists"]
            ],
            "edge_feature_data": edge_feature_data,
            "node_to_graph_idx": torch.tensor(
                list(
                    self.__create_node_to_graph_idx(
                        accumulated_minibatch_data["num_nodes_per_graph"]
                    )
                ),
                dtype=torch.int64,
                device=device,
            ),
            "reference_node_graph_idx": {
                ref_name: torch.tensor(ref_node_graph_idx, dtype=torch.int64, device=device)
                for ref_name, ref_node_graph_idx in accumulated_minibatch_data[
                    "reference_node_graph_idx"
                ].items()
            },
            "reference_node_ids": {
                ref_name: torch.tensor(
                    np.concatenate(ref_node_idxs).astype(np.int32), dtype=torch.int64, device=device
                )
                for ref_name, ref_node_idxs in accumulated_minibatch_data[
                    "reference_node_ids"
                ].items()
            },
            "num_graphs": len(accumulated_minibatch_data["num_nodes_per_graph"]),
        }

    # endregion



from typing_extensions import Final, Literal

import logging
import math
import numpy as np
import torch
import torch.nn as nn
from dpu_utils.codeutils import split_identifier_into_parts
from dpu_utils.mlutils import BpeVocabulary, CharTensorizer, Vocabulary
from typing import Any, Counter, Dict, List, NamedTuple, Optional, Union

from ptgnn.baseneuralmodel import AbstractNeuralModel
from ptgnn.neuralmodels.gnn.structs import AbstractNodeEmbedder


class TokenUnitEmbedder(nn.Module):
    def __init__(self, vocabulary_size: int, embedding_size: int, dropout_rate: float):
        super().__init__()
        self.__embeddings = nn.Embedding(
            num_embeddings=vocabulary_size, embedding_dim=embedding_size
        )
        nn.init.xavier_uniform_(self.__embeddings.weight)  # TODO: Reconsider later?
        self.__dropout_layer = nn.Dropout(p=dropout_rate)

    @property
    def embedding_layer(self) -> nn.Embedding:
        return self.__embeddings

    def forward(self, token_idxs: torch.Tensor) -> torch.Tensor:
        return self.__dropout_layer(self.__embeddings(token_idxs))  # [B, D]


class SubtokenUnitEmbedder(nn.Module):
    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        dropout_rate: float,
        subtoken_combination_kind: str,
        use_dense_output: bool = True,
    ):
        super().__init__()
        assert subtoken_combination_kind in {"mean", "max", "sum"}
        self.__subtoken_combination_kind = subtoken_combination_kind
        self.__embeddings = nn.Embedding(
            num_embeddings=vocabulary_size, embedding_dim=embedding_size
        )
        nn.init.uniform_(self.__embeddings.weight)
        if use_dense_output:
            self.__out_layer = nn.Linear(embedding_size, embedding_size, bias=False)
            nn.init.xavier_uniform_(self.__out_layer.weight)
        else:
            self.__out_layer = None

        self.__dropout_layer = nn.Dropout(p=dropout_rate)

    @property
    def embedding_layer(self) -> nn.Embedding:
        return self.__embeddings

    def forward(self, token_idxs: torch.Tensor, lengths: torch.Tensor) -> torch.Tensor:
        """
        :param token_idxs: The subtoken ids in a [B, max_num_subtokens] matrix.
        :param lengths: A [B]-sized vector containing the lengths
        :return: a [B, D] matrix of D-sized representations, one per input example.
        """
        embedded = self.__embeddings(token_idxs)  # [B, max_num_subtokens, D]
        mask = torch.arange(embedded.shape[1], device=lengths.device).unsqueeze(
            0
        ) < lengths.unsqueeze(
            -1
        )  # [B, max_num_subtokens]

        if self.__subtoken_combination_kind == "mean":
            embedded = embedded * mask.unsqueeze(-1).float()
            embedded = embedded.sum(dim=-2) / (lengths.unsqueeze(-1).float() + 1e-10)  # [B, D]
        elif self.__subtoken_combination_kind == "sum":
            embedded = embedded * mask.unsqueeze(-1).float()
            embedded = embedded.sum(dim=-2)  # [B, D]
        elif self.__subtoken_combination_kind == "max":
            embedded.masked_fill_(mask=~mask.unsqueeze(-1), value=-math.inf)
            embedded, _ = embedded.max(dim=-2)  # [B, D]
        else:
            raise ValueError(
                f'Unrecognized subtoken combination "{self.__subtoken_combination_kind}".'
            )
        if self.__out_layer is not None:
            embedded = self.__out_layer(embedded)
        return self.__dropout_layer(embedded)


class CnnConfig(NamedTuple):
    l1_filters: int
    l1_window_size: int
    l2_filters: int
    l2_window_size: int
    lout_window_size: int


class CharUnitEmbedder(nn.Module):
    def __init__(
        self,
        num_chars: int,
        embedding_size: int,
        config: CnnConfig,
        dropout_rate: float = 0.0,
    ):
        super().__init__()
        self.__num_chars_in_vocabulary = num_chars
        self.__conv_l1 = nn.Conv1d(
            in_channels=num_chars, out_channels=config.l1_filters, kernel_size=config.l1_window_size
        )

        self.__conv_l2 = nn.Conv1d(
            in_channels=config.l1_filters,
            out_channels=config.l2_filters,
            kernel_size=config.l2_window_size,
        )

        self.__conv_l3 = nn.Conv1d(
            in_channels=config.l2_filters,
            out_channels=embedding_size,
            kernel_size=config.lout_window_size,
            bias=False,
        )
        self.__dropout = nn.Dropout(p=dropout_rate)

    def forward(self, chars):
        """
        :param chars: [B, max_num_chars]
        :return: [B, D]
        """
        cnn_input = nn.functional.one_hot(
            chars, self.__num_chars_in_vocabulary
        )  # [B , max_num_chars, char_vocab_size]
        cnn_input = cnn_input.transpose(1, 2).float()
        l1_out = self.__conv_l1(cnn_input)
        l2_out = self.__conv_l2(nn.functional.relu(l1_out))
        l3_out = self.__conv_l3(nn.functional.relu(l2_out))  # [B, Dout, max_num_chars - pad]

        summary, _ = torch.max(l3_out, dim=-1)  # [B, D']
        return self.__dropout(summary)


class StrElementRepresentationModel(
    AbstractNeuralModel[str, Any, Union[TokenUnitEmbedder, SubtokenUnitEmbedder, CharUnitEmbedder]],
    AbstractNodeEmbedder,
):
    """
    A model that accepts strings and returns a single representation (embedding) for each one of them.
    """

    def __init__(
        self,
        *,
        token_splitting: Literal["token", "subtoken", "bpe", "char"],
        embedding_size: int = 128,
        dropout_rate: float = 0.2,
        # Vocabulary Options
        vocabulary_size: int = 10000,
        min_freq_threshold: int = 5,
        # BPE/Subtoken Options
        max_num_subtokens: Optional[int] = 5,
        subtoken_combination: Literal["sum", "mean", "max"] = "sum",
        # Char Models
        cnn_config: CnnConfig = CnnConfig(
            l1_filters=256, l1_window_size=3, l2_filters=128, l2_window_size=3, lout_window_size=3
        ),
        max_num_chars: int = 15,
    ):
        super().__init__()
        self._splitting_kind: Final = token_splitting
        self.embedding_size: Final = embedding_size
        self.dropout_rate: Final = dropout_rate
        self.__vocabulary: Union[Vocabulary, BpeVocabulary, CharTensorizer]
        if token_splitting in {"bpe", "subtoken"}:
            self.max_num_subtokens: Final = max_num_subtokens
            self.subtoken_combination: Final = subtoken_combination
        elif token_splitting == "char":
            self.cnn_config: Final = cnn_config
            self.max_num_chars: Final = max_num_chars
        if token_splitting != "char":
            self.max_vocabulary_size: Final = vocabulary_size
            self.min_freq_threshold: Final = min_freq_threshold

    LOGGER: Final = logging.getLogger(__name__)

    def representation_size(self) -> int:
        return self.embedding_size

    @property
    def splitting_kind(self) -> Literal["token", "subtoken", "bpe", "char"]:
        return self._splitting_kind

    # region Metadata Loading
    def initialize_metadata(self) -> None:
        self.__tok_counter = Counter[str]()

    def update_metadata_from(self, datapoint: str) -> None:
        if self.splitting_kind in {"token", "bpe"}:
            self.__tok_counter[datapoint] += 1
        elif self.splitting_kind == "subtoken":
            self.__tok_counter.update(split_identifier_into_parts(datapoint))
        elif self.splitting_kind == "char":
            pass
        else:
            raise ValueError(f'Unrecognized token splitting method "{self.splitting_kind}".')

    def finalize_metadata(self) -> None:
        if self.splitting_kind in {"token", "subtoken"}:
            self.__vocabulary = Vocabulary.create_vocabulary(
                self.__tok_counter,
                max_size=self.max_vocabulary_size,
                count_threshold=self.min_freq_threshold,
            )
        elif self.splitting_kind == "bpe":
            self.__vocabulary = BpeVocabulary(self.max_vocabulary_size)
            self.__vocabulary.create_vocabulary(self.__tok_counter)
        elif self.splitting_kind == "char":
            self.__vocabulary = CharTensorizer(
                max_num_chars=self.max_num_chars, lower_case_all=False, include_space=False
            )
        else:
            raise ValueError(f'Unrecognized token splitting method "{self.splitting_kind}"')

        del self.__tok_counter

    def build_neural_module(self) -> Union[TokenUnitEmbedder, SubtokenUnitEmbedder]:
        if self.splitting_kind == "token":
            vocabulary_size = len(self.vocabulary)
            embedding_size = self.embedding_size
            return TokenUnitEmbedder(vocabulary_size, embedding_size, self.dropout_rate)
        elif self.splitting_kind in {"bpe", "subtoken"}:
            vocabulary_size = len(self.vocabulary)
            embedding_size = self.embedding_size
            return SubtokenUnitEmbedder(
                vocabulary_size,
                embedding_size,
                self.dropout_rate,
                self.subtoken_combination,
            )
        elif self.splitting_kind == "char":
            return CharUnitEmbedder(
                num_chars=self.vocabulary.num_chars_in_vocabulary(),
                embedding_size=self.embedding_size,
                config=self.cnn_config,
                dropout_rate=self.dropout_rate,
            )
        else:
            raise ValueError(f'Unrecognized token splitting method "{self.splitting_kind}"')

    @property
    def vocabulary(self) -> Union[Vocabulary, BpeVocabulary, CharTensorizer]:
        return self.__vocabulary

    # endregion

    # region Tensorization
    def tensorize(self, datapoint: str, return_str_rep: bool = False):
        if self.splitting_kind == "token":
            token_idxs = self.vocabulary.get_id_or_unk(datapoint)
            str_repr = datapoint
        elif self.splitting_kind == "subtoken":
            subtoks = split_identifier_into_parts(datapoint)
            if len(subtoks) == 0:
                subtoks = [Vocabulary.get_unk()]
            token_idxs = self.vocabulary.get_id_or_unk_multiple(subtoks)
        elif self.splitting_kind == "bpe":
            if len(datapoint) == 0:
                datapoint = "<empty>"
            token_idxs = self.vocabulary.get_id_or_unk_for_text(datapoint)
            if return_str_rep:  # Do _not_ compute for efficiency
                str_repr = self.vocabulary.tokenize(datapoint)
        elif self.splitting_kind == "char":
            token_idxs = self.vocabulary.tensorize_str(datapoint)
            if return_str_rep:
                str_repr = datapoint[: self.vocabulary.max_char_length]
        else:
            raise ValueError(f'Unrecognized token splitting method "{self.splitting_kind}".')

        if return_str_rep:
            return token_idxs, str_repr
        return token_idxs

    # endregion

    # region Minibatching
    def initialize_minibatch(self) -> Dict[str, Any]:
        return {"token_idxs": []}

    def extend_minibatch_with(
        self, tensorized_datapoint: Union[int, List[int]], partial_minibatch: Dict[str, Any]
    ) -> bool:
        partial_minibatch["token_idxs"].append(tensorized_datapoint)
        return True

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        if self.splitting_kind == "token":
            return {
                "token_idxs": torch.tensor(
                    accumulated_minibatch_data["token_idxs"], dtype=torch.int64, device=device
                ),
            }
        elif self.splitting_kind in {"subtoken", "bpe"}:
            max_num_subtokens = max(len(t) for t in accumulated_minibatch_data["token_idxs"])
            if self.max_num_subtokens is not None:
                max_num_subtokens = min(max_num_subtokens, self.max_num_subtokens)

            subtoken_idxs = np.zeros(
                (len(accumulated_minibatch_data["token_idxs"]), max_num_subtokens), dtype=np.int32
            )
            lengths = np.empty(len(accumulated_minibatch_data["token_idxs"]), dtype=np.int32)
            for i, subtokens in enumerate(accumulated_minibatch_data["token_idxs"]):
                idxs = subtokens[:max_num_subtokens]
                subtoken_idxs[i, : len(idxs)] = idxs
                lengths[i] = len(idxs)

            return {
                "token_idxs": torch.tensor(subtoken_idxs, dtype=torch.int64, device=device),
                "lengths": torch.tensor(lengths, dtype=torch.int64, device=device),
            }
        elif self.splitting_kind == "char":
            return {
                "chars": torch.tensor(
                    np.stack(accumulated_minibatch_data["token_idxs"], axis=0),
                    dtype=torch.int64,
                    device=device,
                )
            }
        else:
            raise Exception("Non-reachable state.")

    # endregion



import torch
from torch import nn
from torch_scatter import scatter
from typing import Dict, List, Optional, Tuple, Union

from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import (
    AbstractMessageAggregation,
    AbstractMessagePassingLayer,
)
from ptgnn.neuralmodels.mlp import MLP


class PnaMessageAggregation(AbstractMessageAggregation):
    """
    Principal Neighbourhood Aggregation for Graph Nets

    https://arxiv.org/abs/2004.05718
    """

    def __init__(
        self,
        delta: float = 1,
    ):
        super().__init__()
        self._delta = delta  # See Eq 5 of paper

    def forward(self, messages: torch.Tensor, message_targets: torch.Tensor, num_nodes):
        degree = scatter(
            torch.ones_like(message_targets),
            index=message_targets,
            dim_size=num_nodes,
            reduce="sum",
        )

        sum_agg = scatter(messages, index=message_targets, dim=0, dim_size=num_nodes, reduce="sum")
        mean_agg = sum_agg / (degree.unsqueeze(-1) + 1e-5)
        max_agg = scatter(messages, index=message_targets, dim=0, dim_size=num_nodes, reduce="max")
        min_agg = scatter(messages, index=message_targets, dim=0, dim_size=num_nodes, reduce="min")

        std_components = torch.relu(messages.pow(2) - mean_agg[message_targets].pow(2)) + 1e-10
        std = torch.sqrt(
            scatter(std_components, index=message_targets, dim=0, dim_size=num_nodes, reduce="sum")
        )

        all_aggregations = torch.cat([sum_agg, mean_agg, max_agg, min_agg, std], dim=-1)

        scaler_p1 = torch.log(degree.float() + 1).unsqueeze(-1) / self._delta
        scaler_m1 = 1 / (scaler_p1 + 1e-3)

        return torch.cat(
            [all_aggregations, all_aggregations * scaler_p1, all_aggregations * scaler_m1], dim=-1
        )

    def output_state_size(self, message_input_size: int) -> int:
        return message_input_size * 5 * 3
